"""Plotting functions for geometric shapes."""
from collections import deque
import numpy as np
from mpl_toolkits import mplot3d
import pytransform3d.plot_utils as ppu
import pytransform3d.rotations as pr
import pytransform3d.transformations as pt
from .utils import plane_basis_from_normal


def plot_line(ax, line_point, line_direction, length=10):
    """Plot line.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    length : float, optional (default: 10)
        Length of the line that should be displayed, since we cannot display
        an infinitely long line.
    """
    line = np.vstack([line_point + t * line_direction
                      for t in np.linspace(-length / 2, length / 2, 2)])
    ax.scatter(line_point[0], line_point[1], line_point[2])
    ax.plot(line[:, 0], line[:, 1], line[:, 2])


def plot_segment(ax, segment_start, segment_end, alpha=1.0, c=None, lw=2):
    """Plot line.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    alpha : float, optional (default: 1)
        Alpha for line and points.

    c : str, optional (default: None)
        Color.

    lw : int
        Line width.
    """
    points = np.vstack((segment_start, segment_end))
    ax.scatter(
        points[:, 0], points[:, 1], points[:, 2], alpha=alpha, c=c, lw=lw)
    ax.plot(points[:, 0], points[:, 1], points[:, 2], alpha=alpha, c=c, lw=lw)


def plot_triangle(ax, triangle_points, surface_alpha=0.1):
    """Plot triangle.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    triangle_points : array, shape (3, 3)
        Each row contains a point of the triangle (A, B, C).

    surface_alpha : float, optional (default: 0.1)
        Alpha value of the rectangle surface.
    """
    try:  # Matplotlib < 3.5
        triangle = mplot3d.art3d.Poly3DCollection(triangle_points)
    except ValueError:  # Matplotlib >= 3.5
        triangle = mplot3d.art3d.Poly3DCollection(
            triangle_points.reshape(1, 3, 3))
    triangle.set_alpha(surface_alpha)
    ax.add_collection3d(triangle)

    triangle_points = np.vstack((triangle_points, [triangle_points[0]]))
    ax.plot(triangle_points[:, 0], triangle_points[:, 1], triangle_points[:, 2])


def plot_rectangle(ax, rectangle_center, rectangle_axes, rectangle_lengths,
                   show_axes=True, surface_alpha=0.1):
    """Plot rectangle.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    show_axes : bool, optional (default: True)
        Show axes of the rectangle as arrows.

    surface_alpha : float, optional (default: 0.1)
        Alpha value of the rectangle surface.
    """
    rectangle_points = np.vstack((
        rectangle_center + -0.5 * rectangle_axes[0] * rectangle_lengths[0]
        + -0.5 * rectangle_axes[1] * rectangle_lengths[1],
        rectangle_center + 0.5 * rectangle_axes[0] * rectangle_lengths[0]
        + -0.5 * rectangle_axes[1] * rectangle_lengths[1],
        rectangle_center + 0.5 * rectangle_axes[0] * rectangle_lengths[0]
        + 0.5 * rectangle_axes[1] * rectangle_lengths[1],
        rectangle_center + -0.5 * rectangle_axes[0] * rectangle_lengths[0]
        + 0.5 * rectangle_axes[1] * rectangle_lengths[1],
        rectangle_center + -0.5 * rectangle_axes[0] * rectangle_lengths[0]
        + -0.5 * rectangle_axes[1] * rectangle_lengths[1],
    ))

    vertices = np.vstack((
        rectangle_points[0], rectangle_points[1], rectangle_points[2],
        rectangle_points[0], rectangle_points[2], rectangle_points[3]))
    try:  # Matplotlib < 3.5
        rectangle = mplot3d.art3d.Poly3DCollection(vertices)
    except ValueError:  # Matplotlib >= 3.5
        rectangle = mplot3d.art3d.Poly3DCollection(vertices.reshape(2, 3, 3))
    rectangle.set_alpha(surface_alpha)
    ax.add_collection3d(rectangle)

    ax.plot(rectangle_points[:, 0], rectangle_points[:, 1],
            rectangle_points[:, 2])
    ax.scatter(rectangle_center[0], rectangle_center[1], rectangle_center[2])
    if show_axes:
        ppu.plot_vector(
            ax=ax, start=rectangle_center, direction=rectangle_axes[0],
            s=0.5 * rectangle_lengths[0], color="r")
        ppu.plot_vector(
            ax=ax, start=rectangle_center, direction=rectangle_axes[1],
            s=0.5 * rectangle_lengths[1], color="g")


def plot_circle(ax, center, radius, normal, show_normal=False,
                surface_alpha=0.1, color="b"):
    """Plot circle.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    center : array, shape (3,)
        Center of the circle.

    radius : float
        Radius of the circle.

    normal : array, shape (3,)
        Normal to the plane in which the circle lies.

    show_normal : bool, optional (default: False)
        Display normal of the circle plane.

    surface_alpha : float, optional (default: 0.1)
        Alpha value of the circle surface.

    color : str
        Color of surface.
    """
    ax.scatter(center[0], center[1], center[2])
    u, v = plane_basis_from_normal(normal)
    R = np.column_stack((u, v, normal))
    circle = np.array([
        center + R.dot(pr.matrix_from_angle(2, angle)).dot([radius, 0, 0])
        for angle in np.linspace(0, 2 * np.pi, 20)])
    ax.plot(circle[:, 0], circle[:, 1], circle[:, 2])

    vertices = np.array([
        [circle[i], circle[j], center]
        for i, j in zip(range(len(circle) - 1), range(1, len(circle)))])
    try:  # Matplotlib < 3.5
        surface = mplot3d.art3d.Poly3DCollection(vertices)
    except ValueError:  # Matplotlib >= 3.5
        surface = mplot3d.art3d.Poly3DCollection(vertices.reshape(-1, 3, 3))
    surface.set_facecolor(color)
    surface.set_alpha(surface_alpha)
    ax.add_collection3d(surface)

    if show_normal:
        ppu.plot_vector(ax=ax, start=center, direction=normal, s=1.0)


def plot_aabb(ax, mins, maxs, alpha=1.0, color="r"):
    """Plot axis-aligned bounding box.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    mins : array, shape (3,)
        Minimum values along axes.

    maxs : array, shape (3,)
        Maximum values along axes.

    alpha : float, optional (default: 1)
        Alpha value of edges.

    color : str
        Color of edges.
    """
    box2origin = np.eye(4)
    box2origin[:3, 3] = 0.5 * (mins + maxs)
    size = maxs - mins
    ppu.plot_box(ax=ax, A2B=box2origin, size=size, wireframe=True, alpha=alpha, color=color)


def plot_convex(ax, mesh2origin, vertices, triangles, alpha=0.5, color="b"):
    """Plot convex mesh.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_points, 3)
        Vertices of the convex mesh.

    triangles : array, shape (n_triangles, 3)
        Vertex indices of faces.

    alpha : float, optional (default: 0.5)
        Alpha value of faces.

    color : str
        Color of faces.
    """
    points = pt.transform(mesh2origin, pt.vectors_to_points(vertices))[:, :3]
    ax.scatter(points[:, 0], points[:, 1], points[:, 2], color="red")
    faces = np.array([points[[i, j, k]] for i, j, k in triangles])
    surface = mplot3d.art3d.Poly3DCollection(faces)
    surface.set_facecolor(color)
    surface.set_alpha(alpha)
    ax.add_collection3d(surface)
    wireframe = mplot3d.art3d.Line3DCollection(faces)
    ax.add_collection3d(wireframe)


def plot_tetrahedron(ax, vertices, show_triangles=False):
    """Plot tetrahedron.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    vertices : array, shape (4, 3)
        Vertices of the tetrahedron.

    show_triangles : bool, optional (default: False)
        Show edges of triangles in different color.
    """
    line = np.array([
        vertices[0], vertices[1], vertices[2], vertices[0],
        vertices[0], vertices[1], vertices[3], vertices[0],
        vertices[0], vertices[2], vertices[3], vertices[0],
        vertices[1], vertices[2], vertices[3], vertices[1],
    ])
    ax.plot(line[:, 0], line[:, 1], line[:, 2])
    if show_triangles:
        for triangle_index in range(0, len(line), 4):
            for vertex_index, c in zip(range(3), "rgb"):
                s = slice(triangle_index + vertex_index, triangle_index + vertex_index + 2)
                ax.plot(line[s][:, 0], line[s][:, 1], line[s][:, 2], c=c)


def plot_aabb_tree(ax, tree, alpha=0.5, color="red"):
    """Plot tree of axis-aligned bounding boxes.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    tree : aabbtree.AABBTree
        Tree of axis-aligned bounding boxes.

    alpha : float, optional (default: 0.5)
        Alpha value of edges.

    color : str
        Color of edges.
    """
    nodes = deque()
    nodes.append(tree)
    while nodes:
        node = nodes.popleft()
        mins, maxs = np.array(node.aabb.limits).T
        plot_aabb(ax, mins, maxs, alpha=alpha, color=color)
        if not node.is_leaf:
            nodes.extend([node.left, node.right])


def plot_plane(
        ax, plane_point, plane_normal, s=1.0, surface_alpha=0.1, color="b"):
    """Plot rectangle.

    Parameters
    ----------
    ax : Matplotlib 3d axis
        A matplotlib 3d axis.

    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    s : float, optional (default: 1)
        Scaling of the plane that will be drawn.

    surface_alpha : float, optional (default: 0.1)
        Alpha value of the rectangle surface.

    color : str
        Color of faces.
    """
    x_axis, y_axis = plane_basis_from_normal(plane_normal)
    vertices = np.array([
        plane_point + s * x_axis + s * y_axis,
        plane_point - s * x_axis + s * y_axis,
        plane_point + s * x_axis - s * y_axis,
        plane_point - s * x_axis - s * y_axis,
    ])
    vertices = vertices[np.array([0, 1, 2, 1, 3, 2, 2, 1, 0, 2, 3, 1])]

    try:  # Matplotlib < 3.5
        rectangle = mplot3d.art3d.Poly3DCollection(vertices)
    except ValueError:  # Matplotlib >= 3.5
        rectangle = mplot3d.art3d.Poly3DCollection(vertices.reshape(4, 3, 3))
    rectangle.set_facecolor(color)
    rectangle.set_alpha(surface_alpha)
    ax.add_collection3d(rectangle)

    ppu.plot_vector(
        ax=ax, start=plane_point, direction=plane_normal, s=s, color="r")



"""Tools for geometric computations."""
import math
from itertools import product
import numba
import numpy as np
from .utils import norm_vector, transform_point, plane_basis_from_normal


def convert_rectangle_to_segment(rectangle_center, rectangle_extents, i0, i1):
    """Extract line segment from rectangle.

    Parameters
    ----------
    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_extents : array, shape (3, 2)
        Extents along axes of the rectangles:
        0.5 * rectangle_sizes * rectangle_axes.

    i0 : int
        Either 0 or 1, selecting line segment.

    i1 : int
        Either 0 or 1, selecting line segment.

    Returns
    -------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.
    """
    segment_middle = rectangle_center + (2 * i0 - 1) * rectangle_extents[i1]
    segment_start = segment_middle - rectangle_extents[1 - i1]
    segment_end = segment_middle + rectangle_extents[1 - i1]
    return segment_end, segment_start


RECTANGLE_COORDS = np.array([
    [-0.5, -0.5], [-0.5, 0.5], [0.5, -0.5], [0.5, 0.5]])


def convert_rectangle_to_vertices(
        rectangle_center, rectangle_axes, rectangle_lengths):
    """Convert rectangle to vertices.

    Parameters
    ----------
    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    Returns
    -------
    rectangle_points : array, shape (4, 3)
        Vertices of the rectangle.
    """
    return rectangle_center + (RECTANGLE_COORDS * rectangle_lengths).dot(rectangle_axes)


def convert_box_to_face(box2origin, size, i, sign):
    """Convert box to face.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    i : int
        Index of the axis along which we select the face.

    sign : int
        Indicate the direction along the axis.

    Returns
    -------
    face_center : array, shape (3,)
        Center point of the rectangle.

    face_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    face_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.
    """
    other_indices = [0, 1, 2]
    other_indices.remove(i)
    face_center = box2origin[:3, 3] + sign * 0.5 * size[i] * box2origin[:3, i]
    face_axes = np.array([box2origin[:3, j] for j in other_indices])
    face_lengths = np.array([size[j] for j in other_indices])
    return face_center, face_axes, face_lengths


@numba.njit(cache=True)
def convert_segment_to_line(segment_start, segment_end):
    """Convert line segment to line.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    Returns
    -------
    segment_direction : array, shape (3,)
        Line direction with unit length (or 0).

    segment_length : float
        Length of the line segment.
    """
    segment_direction = segment_end - segment_start
    segment_length = np.linalg.norm(segment_direction)
    if segment_length > 0:
        segment_direction /= segment_length
    return segment_direction, segment_length


BOX_COORDS = np.array(list(product([-0.5, 0.5], repeat=3)))


@numba.njit(
    numba.float64[:, :](numba.float64[:, ::1], numba.float64[::1]),
    cache=True)
def convert_box_to_vertices(box2origin, size):
    """Convert box to vertices.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    Returns
    -------
    box_points : array, shape (8, 3)
        Vertices of the box.
    """
    return box2origin[:3, 3] + (BOX_COORDS * size).dot(box2origin[:3, :3].T)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[:, ::1], numba.float64,
                     numba.float64),
    cache=True)
def support_function_cylinder(
        search_direction, cylinder2origin, radius, length):
    """Compute extreme point of cylinder along a direction.

    You can find similar implementations here:

    * https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/Collider.h#L42
      (Copyright (c) 2017 Kevin Moran, MIT License or Unlicense)
    * https://github.com/bulletphysics/bullet3/blob/e306b274f1885f32b7e9d65062aa942b398805c2/src/BulletCollision/CollisionShapes/btConvexShape.cpp#L167
      (Copyright (c) 2003-2009 Erwin Coumans, zlib license)

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Radius of the cylinder.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    local_dir = np.dot(cylinder2origin[:3, :3].T, search_direction)

    s = math.sqrt(local_dir[0] * local_dir[0] + local_dir[1] * local_dir[1])
    if local_dir[2] < 0.0:
        z = -0.5 * length
    else:
        z = 0.5 * length
    if s == 0.0:
        local_vertex = np.array([radius, 0.0, z])
    else:
        d = radius / s
        local_vertex = np.array([local_dir[0] * d, local_dir[1] * d, z])
    return transform_point(cylinder2origin, local_vertex)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[:, ::1], numba.float64,
                     numba.float64),
    cache=True)
def support_function_capsule(
        search_direction, capsule2origin, radius, height):
    """Compute extreme point of cylinder along a direction.

    You can find similar implementations here:

    * https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/Collider.h#L57
      (Copyright (c) 2017 Kevin Moran, MIT License or Unlicense)
    * https://github.com/bulletphysics/bullet3/blob/e306b274f1885f32b7e9d65062aa942b398805c2/src/BulletCollision/CollisionShapes/btConvexShape.cpp#L228
      (Copyright (c) 2003-2009 Erwin Coumans, zlib license)

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the cylinder.

    height : float
        Height of the cylinder.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    local_dir = np.dot(capsule2origin[:3, :3].T, search_direction)

    s = math.sqrt(local_dir[0] * local_dir[0] + local_dir[1] * local_dir[1]
                  + local_dir[2] * local_dir[2])
    if s == 0.0:
        local_vertex = np.array([radius, 0, 0])
    else:
        local_vertex = local_dir * (radius / s)
    if local_dir[2] > 0.0:
        local_vertex[2] += 0.5 * height
    else:
        local_vertex[2] -= 0.5 * height

    return transform_point(capsule2origin, local_vertex)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[:, ::1], numba.float64[::1]),
    cache=True)
def support_function_ellipsoid(
        search_direction, ellipsoid2origin, radii):
    """Compute extreme point of ellipsoid along a direction.

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    local_dir = np.dot(ellipsoid2origin[:3, :3].T, search_direction)
    local_vertex = norm_vector(local_dir * radii) * radii
    return transform_point(ellipsoid2origin, local_vertex)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[:, ::1], numba.float64[::1]),
    cache=True)
def support_function_box(search_direction, box2origin, half_lengths):
    """Compute extreme point of box along a direction.

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    box2origin : array, shape (4, 4)
        Pose of the box.

    half_lengths : array, shape (3,)
        Half lengths of the box along its axes.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    local_dir = np.dot(box2origin[:3, :3].T, search_direction)
    local_vertex = np.sign(local_dir) * half_lengths
    return transform_point(box2origin, local_vertex)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[::1], numba.float64),
    cache=True)
def support_function_sphere(search_direction, center, radius):
    """Compute extreme point of box along a direction.

    You can find similar implementations here:

    * https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/Collider.h#L33
      (Copyright (c) 2017 Kevin Moran, MIT License or Unlicense)

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    s_norm = np.linalg.norm(search_direction)
    if s_norm == 0.0:
        vertex = center + np.array([0, 0, radius])
    else:
        vertex = center + search_direction / s_norm * radius
    return vertex


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[::1], numba.float64,
                     numba.float64[::1]),
    cache=True)
def support_function_disk(search_direction, center, radius, normal):
    """Compute extreme point of disk along a direction.

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    center : array, shape (3,)
        Center of the disk.

    radius : float
        Radius of the disk.

    normal : array, shape (3,)
        Normal to the plane in which the disk lies.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    x, y = plane_basis_from_normal(normal)
    R = np.column_stack((x, y, normal))
    point = np.dot(R.T, search_direction)
    point[2] = 0.0
    norm = np.linalg.norm(point)
    if norm == 0.0:
        return np.copy(center)
    point *= radius / norm
    return center + np.dot(R, point)


@numba.njit(
    numba.float64[:](numba.float64[::1], numba.float64[:, ::1], numba.float64,
                     numba.float64),
    cache=True)
def support_function_cone(search_direction, cone2origin, radius, height):
    """Compute extreme point of cone along a direction.

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction.

    cone2origin : array, shape (4, 4)
        Pose of the cone.

    radius : float
        Radius of the cone.

    height : float
        Length of the cone.

    Returns
    -------
    extreme_point : array, shape (3,)
        Extreme point along search direction.
    """
    local_dir = np.dot(cone2origin[:3, :3].T, search_direction)
    disk_point = np.array([local_dir[0], local_dir[1], 0.0])
    norm = np.linalg.norm(disk_point)
    if norm == 0.0:
        disk_point = np.array([0.0, 0.0, 0.0])
    else:
        disk_point *= radius / norm
    if np.dot(local_dir, disk_point) >= local_dir[2] * height:
        point_in_cone = disk_point
    else:
        point_in_cone = np.array([0.0, 0.0, height])
    return transform_point(cone2origin, point_in_cone)


@numba.njit(cache=True)
def hesse_normal_form(plane_point, plane_normal):
    """Computes Hesse normal form of a plane.

    In the Hesse normal form (x * n - d = 0), x is any point on the plane,
    n is the plane's normal, and d ist the distance from the origin to the
    plane along its normal.

    Parameters
    ----------
    plane_point : array, shape (3,)
        Point on the plane.

    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    Returns
    -------
    plane_normal : array, shape (3,)
        Normal of the plane. We assume unit length.

    d : float, optional (default: None)
        Distance of the plane to origin in Hesse normal form.
    """
    return plane_normal, np.dot(plane_point, plane_normal)


def line_from_pluecker(line_direction, line_moment):
    """Computes line from Plücker coordinates.

    Parameters
    ----------
    line_direction : array, shape (3,)
        Direction of the line. Not necessarily of unit length.

    line_moment : array, shape (3,)
        Moment of the line.

    Returns
    -------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.
    """
    line_point = np.cross(line_direction, line_moment)
    line_direction = line_direction
    line_dir_norm_squared = np.dot(line_direction, line_direction)
    if line_dir_norm_squared > 0.0:
        line_point /= line_dir_norm_squared
        line_direction /= math.sqrt(line_dir_norm_squared)
    return line_point, line_direction



"""Gilbert-Johnson-Keerthi (GJK) for distance calculation of convex shapes."""
from .gjk_distance import gjk, gjk_with_simplex
from .gjk_intersection import gjk_intersection


__all__ = ["gjk", "gjk_with_simplex", "gjk_intersection"]



"""Utility functions."""
import math
import numba
import numpy as np


MAX_FLOAT = np.finfo(float).max


@numba.njit(numba.float64[::1](numba.float64[::1]), cache=True)
def norm_vector(v):
    """Normalize vector.

    Parameters
    ----------
    v : array, shape (n,)
        nd vector

    Returns
    -------
    u : array, shape (n,)
        nd unit vector with norm 1 or the zero vector
    """
    norm = np.linalg.norm(v)
    if norm == 0.0:
        return v

    return v / norm


@numba.njit(numba.types.Tuple(
    (numba.float64[::1], numba.float64[::1]))(numba.float64[::1]), cache=True)
def plane_basis_from_normal(plane_normal):
    """Compute two basis vectors of a plane from the plane's normal vector.

    Note that there are infinitely many solutions because any rotation of the
    basis vectors about the normal is also a solution. This function
    deterministically picks one of the solutions.

    The two basis vectors of the plane together with the normal form an
    orthonormal basis in 3D space and could be used as columns to form a
    rotation matrix.

    Parameters
    ----------
    plane_normal : array-like, shape (3,)
        Plane normal of unit length.

    Returns
    -------
    x_axis : array, shape (3,)
        x-axis of the plane.

    y_axis : array, shape (3,)
        y-axis of the plane.
    """
    if abs(plane_normal[0]) >= abs(plane_normal[1]):
        # x or z is the largest magnitude component, swap them
        length = math.sqrt(
            plane_normal[0] * plane_normal[0]
            + plane_normal[2] * plane_normal[2])
        x_axis = np.array([-plane_normal[2] / length, 0.0,
                           plane_normal[0] / length])
        y_axis = np.array([
            plane_normal[1] * x_axis[2],
            plane_normal[2] * x_axis[0] - plane_normal[0] * x_axis[2],
            -plane_normal[1] * x_axis[0]])
    else:
        # y or z is the largest magnitude component, swap them
        length = math.sqrt(plane_normal[1] * plane_normal[1]
                           + plane_normal[2] * plane_normal[2])
        x_axis = np.array([0.0, plane_normal[2] / length,
                           -plane_normal[1] / length])
        y_axis = np.array([
            plane_normal[1] * x_axis[2] - plane_normal[2] * x_axis[1],
            -plane_normal[0] * x_axis[2], plane_normal[0] * x_axis[1]])
    return x_axis, y_axis


@numba.njit(numba.float64[::1](numba.float64[:, ::1], numba.float64[::1]),
            cache=True)
def transform_point(A2B, point_in_A):
    """Transform a point from frame A to frame B.

    Parameters
    ----------
    A2B : array, shape (4, 4)
        Transform from frame A to frame B as homogeneous matrix.

    point_in_A : array, shape (3,)
        Point in frame A.

    Returns
    -------
    point_in_B : array, shape (3,)
        Point in frame B.
    """
    return A2B[:3, 3] + np.dot(A2B[:3, :3], point_in_A)


def angles_between_vectors(A, B):
    """Compute angle between two vectors.

    Parameters
    ----------
    A : array, shape (n_vectors, 3)
        3d vectors

    B : array-like, shape (n_vectors, 3)
        3d vectors

    Returns
    -------
    angles : array, shape (n_vectors,)
        Angles between pairs of vectors from A and B
    """
    return np.arccos(
        np.clip(np.sum(A * B, axis=1) / (np.linalg.norm(A, axis=1)
                                         * np.linalg.norm(B, axis=1)),
                -1.0, 1.0))


EPSILON = np.finfo(float).eps



import math
from collections import namedtuple
import numpy as np
import pytransform3d.rotations as pr
from ..utils import norm_vector
from ..geometry import convert_segment_to_line


def point_to_circle(point, center, radius, normal, epsilon=1e-6):
    """Compute the shortest distance between point and circle (only line).

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    center : array, shape (3,)
        Center of the circle.

    radius : float
        Radius of the circle.

    normal : array, shape (3,)
        Normal to the plane in which the circle lies.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between point and circle.

    closest_point_circle : array, shape (3,)
        Closest point on the circle.
    """
    # signed distance from point to plane of circle
    diff = point - center
    dist_to_plane = diff.dot(normal)

    # projection of P - C onto plane is Q - C = P - C - dist_to_plane * normal
    diff_in_plane = diff - dist_to_plane * normal
    sqr_len = diff_in_plane.dot(diff_in_plane)

    if sqr_len >= epsilon:
        closest_point_circle = (
            center + (radius / math.sqrt(sqr_len)) * diff_in_plane)
        dist = np.linalg.norm(point - closest_point_circle)
    else:  # on the line defined by center and normal of the circle
        plane_direction = norm_vector(pr.perpendicular_to_vector(normal))
        closest_point_circle = center + radius * plane_direction
        dist = math.sqrt(radius * radius + dist_to_plane * dist_to_plane)

    return dist, closest_point_circle


def line_to_circle(line_point, line_direction, center, radius, normal):
    """Compute the shortest distance between line and circle.

    This is the nonpolynomial-based algorithm that uses bisection.
    More details here:

    * https://www.geometrictools.com/Documentation/DistanceToCircle3.pdf
    * https://www.geometrictools.com/GTE/Mathematics/DistLine3Circle3.h

    Adapted from C++ implementation of
    David Eberly, Geometric Tools, Redmond WA 98052 (Copyright (c) 1998-2022)
    Distributed under the Boost Software License, Version 1.0.

    * https://www.boost.org/LICENSE_1_0.txt
    * https://www.geometrictools.com/License/Boost/LICENSE_1_0.txt

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    center : array, shape (3,)
        Center of the circle.

    radius : float
        Radius of the circle.

    normal : array, shape (3,)
        Normal to the plane in which the circle lies.

    Returns
    -------
    dist : float
        The shortest distance between line and circle.

    closest_point_line : array, shape (3,)
        Closest point on line.

    closest_point_circle : array, shape (3,)
        Closest point on circle.
    """
    # The line is P(t) == line_point + t * line_direction.
    # The circle is |x - center| == radius with <normal, x - center> == 0.

    # We transform the line to the circle coordinate frame, in which the circle
    # center is at the origin (0, 0, 0) (and the normal is (0, 0, 1)).
    line_point = line_point - center

    line_direction_cross_normal = np.cross(line_direction, normal)
    line_point_cross_normal = np.cross(line_point, normal)

    m0_squared = np.dot(
        line_direction_cross_normal, line_direction_cross_normal)
    if m0_squared > 0.0:
        closest_point_line, closest_point_circle = _case_line_and_normal_not_parallel(
            line_point, line_direction, center, radius, normal,
            m0_squared, line_direction_cross_normal,
            line_point_cross_normal)
    else:  # Line direction and the plane normal are parallel.
        closest_point_line, closest_point_circle = _case_line_and_normal_parallel(
            line_point, line_direction, center, radius, normal,
            line_point_cross_normal)

    dist = np.linalg.norm(closest_point_line - closest_point_circle)
    return dist, closest_point_line, closest_point_circle


def _case_line_and_normal_not_parallel(
        line_point, line_direction, center, radius, normal, m0_squared,
        line_direction_cross_normal, line_point_cross_normal):
    # Compute the critical points s for F'(s) = 0.
    # The line direction M and the plane normal N are not parallel.
    # Move the line_point = (b0, b1, b2) to
    # line_point' = line_point + lmbda * line_direction = (0, b1', b2').
    sin_directions = math.sqrt(m0_squared)
    lmbda = -np.dot(line_direction_cross_normal,
                    line_point_cross_normal) / m0_squared
    old_line_point = line_point
    line_point = old_line_point + lmbda * line_direction
    line_point_cross_normal += lmbda * line_direction_cross_normal
    m2b2 = np.dot(line_direction, line_point)
    b1_squared = np.dot(line_point_cross_normal, line_point_cross_normal)

    if b1_squared > 0.0:
        roots = _case_general(
            b1_squared, m0_squared, m2b2, radius, sin_directions)
    else:
        roots = _case_b1_is_zero(m2b2, radius * sin_directions)

    candidates = [
        _convert_root_to_candidate(
            root, lmbda, old_line_point, line_direction, center, radius, normal)
        for root in roots]
    candidate = candidates[np.argmin([c.dist_squared for c in candidates])]
    return candidate.closest_point_line, candidate.closest_point_circle


def _case_general(b1_squared, m0_squared, m2b2, radius, sin_directions):
    # line_point' = (0, b1', b2') where b1' != 0. See Sections 1.1.2
    # and 1.2.2 of the PDF documentation.
    roots = []
    b1 = math.sqrt(b1_squared)
    radius_m0_squared = radius * m0_squared
    radius_sin_directions = radius * sin_directions
    if radius_m0_squared > b1:
        s_hat2 = abs(m0_squared * b1_squared ** (2.0 / 3.0) - b1_squared)  # TODO why can this be negative?
        s_hat = math.sqrt(s_hat2) / sin_directions
        g_hat = radius_m0_squared * s_hat / math.sqrt(
            m0_squared * s_hat * s_hat + b1_squared)
        cutoff = g_hat - s_hat
        if m2b2 <= -cutoff:
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared, -m2b2,
                -m2b2 + radius_sin_directions))
            if m2b2 == -cutoff:
                roots.append(-s_hat)
        elif m2b2 >= cutoff:
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared,
                -m2b2 - radius_sin_directions, -m2b2))
            if m2b2 == cutoff:
                roots.append(s_hat)
        elif m2b2 <= 0.0:
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared, -m2b2,
                -m2b2 + radius_sin_directions))
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared,
                -m2b2 - radius_sin_directions, -s_hat))
        else:
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared,
                -m2b2 - radius_sin_directions, -m2b2))
            roots.append(_bisect_line_circle(
                m2b2, radius_m0_squared, m0_squared, b1_squared, s_hat,
                -m2b2 + radius_sin_directions))
    elif m2b2 < 0.0:
        roots.append(_bisect_line_circle(
            m2b2, radius_m0_squared, m0_squared, b1_squared, -m2b2,
            -m2b2 + radius_sin_directions))
    elif m2b2 > 0.0:
        roots.append(_bisect_line_circle(
            m2b2, radius_m0_squared, m0_squared, b1_squared,
            -m2b2 - radius_sin_directions, -m2b2))
    else:
        roots.append(0.0)

    return roots


def _bisect_line_circle(
        m2b2, rm0_squared, m0_squared, b1_squared,
        s_min, s_max):
    # Bisect the function
    #   f(s) = s + m2b2 - radius * m0_squared * s / sqrt(
    #       m0_squared * s * s + b1sqr)
    # on the specified interval [s_min, s_max].
    for i in range(2, 10):
        s = 0.5 * (s_min + s_max)
        if s + m2b2 - rm0_squared * s / math.sqrt(
                m0_squared * s * s + b1_squared) > 0.0:
            s_max = s
        else:
            s_min = s
    return s


def _case_b1_is_zero(m2b2, radius_sin_directions):
    # The new line origin is B' = (0, 0, b2').
    roots = []
    if m2b2 < 0.0:
        roots.append(-m2b2 + radius_sin_directions)
    elif m2b2 > 0.0:
        roots.append(-m2b2 - radius_sin_directions)
    else:
        roots.append(-m2b2 + radius_sin_directions)
        roots.append(-m2b2 - radius_sin_directions)
    return roots


def _convert_root_to_candidate(
        root, lmbda, old_line_point, line_direction, center, radius, normal):
    t = root + lmbda
    normal_cross_delta = np.cross(
        normal, old_line_point + t * line_direction)
    if any(normal_cross_delta != 0.0):
        closest_point_line, closest_point_circle = _line_circle_closest_points(
            old_line_point, line_direction, center, radius, normal, t)
    else:
        u = pr.perpendicular_to_vector(normal)
        closest_point_line = center
        closest_point_circle = center + radius * u
    diff = closest_point_line - closest_point_circle
    dist_squared = np.dot(diff, diff)
    candidate = LineToCircleCandidate(
        closest_point_line, closest_point_circle, dist_squared)
    return candidate


def _line_circle_closest_points(
        line_point, line_direction, center, radius, normal, t):
    delta = line_point + t * line_direction
    line_closest = center + delta
    delta -= np.dot(normal, delta) * normal
    delta = norm_vector(delta)
    circle_closest = center + radius * delta
    return line_closest, circle_closest


LineToCircleCandidate = namedtuple(
    "LineToCircleCandidate",
    "closest_point_line closest_point_circle dist_squared")


def _case_line_and_normal_parallel(
        line_point, line_direction, center, radius, normal,
        line_point_cross_normal):
    if any(line_point_cross_normal != 0.0):
        # The line is A + t * normal but with A != center.
        closest_point_line, closest_point_circle = _line_circle_closest_points(
            line_point, line_direction, center, radius, normal,
            -np.dot(line_direction, line_point))
    else:
        # The line is center + t * normal, so the circle center is the
        # closest point for the line and all circle points are equidistant
        # from it.
        u = pr.perpendicular_to_vector(normal)
        closest_point_line = center
        closest_point_circle = center + radius * u
    return closest_point_line, closest_point_circle


def line_segment_to_circle(segment_start, segment_end, center, radius, normal):
    """Compute the shortest distance between line segment and circle.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    center : array, shape (3,)
        Center of the circle.

    radius : float
        Radius of the circle.

    normal : array, shape (3,)
        Normal to the plane in which the circle lies.

    Returns
    -------
    dist : float
        The shortest distance between line segment and circle.

    closest_point_line_segment : array, shape (3,)
        Closest point on line segment.

    closest_point_circle : array, shape (3,)
        Closest point on circle.
    """
    return _line_segment_to_circle(
        segment_start, segment_end, center, radius, normal)[:3]


def _line_segment_to_circle(segment_start, segment_end, center, radius, normal):
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)

    dist, closest_point_segment, closest_point_circle = line_to_circle(
        segment_start, segment_direction, center, radius, normal)

    comparison_dimensions = np.where(segment_direction != 0.0)[0]
    assert len(comparison_dimensions) > 0
    comparison_dimension = comparison_dimensions[0]
    t = ((closest_point_segment[comparison_dimension]
          - segment_start[comparison_dimension])
         / segment_direction[comparison_dimension])

    if t < 0.0:
        dist, closest_point_circle = point_to_circle(
            segment_start, center, radius, normal)
        closest_point_segment = segment_start
        on_line = False
    elif t > segment_length:
        dist, closest_point_circle = point_to_circle(
            segment_end, center, radius, normal)
        closest_point_segment = segment_end
        on_line = False
    else:
        on_line = True

    return dist, closest_point_segment, closest_point_circle, on_line



import numpy as np
import pytransform3d.transformations as pt


def point_to_ellipsoid(
        point, ellipsoid2origin, radii, distance_to_surface=False,
        epsilon=1e-16, max_iter=64, check=False):
    """Compute the shortest distance between point and ellipsoid.

    Implementation adapted from 3D Game Engine Design by David H. Eberly.

    Geometric Tools, Inc.
    http://www.geometrictools.com
    Copyright (c) 1998-2006.  All Rights Reserved

    The Wild Magic Version 4 Foundation Library source code is supplied
    under the terms of the license agreement
    (http://www.geometrictools.com/License/Wm4FoundationLicense.pdf)
    and may not be copied or disclosed except in accordance with the terms
    of that agreement.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    distance_to_surface : bool, optional (default: False)
        Compute distance to surface or volume otherwise.

    epsilon : float, optional (default: 1e-16)
        Values smaller than epsilon are considered to be 0.

    max_iter : int, optional (default: 64)
        Maximum number of iterations of the optimization.

    check : bool, optional (default: True)
        Check if transformation matrix is valid before inversion.

    Returns
    -------
    dist : float
        Shortest distance.

    closest_point_ellipsoid : array, shape (3,)
        Closest point on ellipsoid.
    """
    # compute coordinates of point in ellipsoid coordinate system
    origin2ellipsoid = pt.invert_transform(ellipsoid2origin, check=check)
    point_in_ellipsoid = origin2ellipsoid[:3, 3] + origin2ellipsoid[:3, :3].dot(point)

    radii2 = radii ** 2
    point2 = point_in_ellipsoid ** 2
    radii2point2 = radii2 * point2

    # initial guess
    normalized_point_norm = np.linalg.norm(point_in_ellipsoid / radii)
    if normalized_point_norm < 1.0:
        if distance_to_surface:
            if normalized_point_norm < 100.0 * epsilon:
                # Point is the center of the ellipsoid, we have two possible
                # solutions on opposite sides and must select one.
                min_radius_idx = np.argmin(radii)
                closest_point = (
                    ellipsoid2origin[:3, 3]
                    + radii[min_radius_idx] * ellipsoid2origin[:3, min_radius_idx])
                return radii[min_radius_idx], closest_point
            t = 0.0
        else:
            return 0.0, point
    else:
        t = max(radii) * np.linalg.norm(point_in_ellipsoid)

    # Newton's method
    for i in range(max_iter):
        pqr = t + radii2
        pqr2 = pqr ** 2
        s = (pqr2[0] * pqr2[1] * pqr2[2]
             - radii2point2[0] * pqr2[1] * pqr2[2]
             - radii2point2[1] * pqr2[0] * pqr2[2]
             - radii2point2[2] * pqr2[0] * pqr2[1])
        if abs(s) < epsilon:
            break

        pq = pqr[0] * pqr[1]
        pr = pqr[0] * pqr[2]
        qr = pqr[1] * pqr[2]
        pqr_ = pqr[0] * pqr[1] * pqr[2]
        ds = 2.0 * (pqr_ * (qr + pr + pq)
                    - radii2point2[0] * qr * (pqr[1] + pqr[2])
                    - radii2point2[1] * pr * (pqr[0] + pqr[2])
                    - radii2point2[2] * pq * (pqr[0] + pqr[1]))
        t -= s / ds

    closest_point_in_ellipsoid = radii2 * point_in_ellipsoid / pqr
    diff = closest_point_in_ellipsoid - point_in_ellipsoid

    closest_point_ellipsoid = (
        ellipsoid2origin[:3, 3]
        + ellipsoid2origin[:3, :3].dot(closest_point_in_ellipsoid))

    return np.linalg.norm(diff), closest_point_ellipsoid



"""Tools for convex meshes."""
import numba
import numpy as np
from scipy.spatial import ConvexHull

from .utils import angles_between_vectors


class MeshHillClimbingSupportFunction:
    """Mesh support function with hill climbing.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.
    """
    def __init__(self, mesh2origin, vertices, triangles):
        self.mesh2origin = mesh2origin
        self.vertices = vertices

        self.first_idx = np.min(triangles)

        connections = {}
        for i, j, k in triangles:
            if i not in connections:
                connections[i] = set()
            if j not in connections:
                connections[j] = set()
            if k not in connections:
                connections[k] = set()
            connections[i].update((j, k))
            connections[j].update((i, k))
            connections[k].update((i, j))

        self.shortcut_connections = np.array([
            np.argmax(self.vertices[:, 0]), np.argmax(self.vertices[:, 1]),
            np.argmax(self.vertices[:, 2]), np.argmin(self.vertices[:, 0]),
            np.argmin(self.vertices[:, 1]), np.argmin(self.vertices[:, 2])])

        self.connections = numba.typed.Dict.empty(numba.int64, numba.int64[:])
        for idx, connected_indices in connections.items():
            self.connections[idx] = np.fromiter(
                connected_indices, dtype=int, count=len(connected_indices))

    def update_pose(self, mesh2origin):
        """Update pose.

        Parameters
        ----------
        mesh2origin : array, shape (4, 4)
            New pose of the mesh.
        """
        self.mesh2origin = mesh2origin

    def __call__(self, search_direction):
        """Support function.

        Parameters
        ----------
        search_direction : array, shape (3,)
            Search direction.

        Returns
        -------
        idx : int
            Index of support point.

        support_point : array, shape (3,)
            Support point.
        """
        search_direction_in_mesh = np.dot(
            self.mesh2origin[:3, :3].T, search_direction)
        idx = hill_climb_mesh_extreme(
            search_direction_in_mesh, self.first_idx, self.vertices,
            self.connections, self.shortcut_connections)
        self.first_idx = idx  # vertex caching
        return idx, self.mesh2origin[:3, 3] + np.dot(
            self.mesh2origin[:3, :3], self.vertices[idx])


@numba.njit(
    numba.int64(numba.float64[:], numba.int64, numba.float64[:, :],
                numba.types.DictType(numba.int64, numba.int64[:]),
                numba.optional(numba.int64[:])),
    cache=True)
def hill_climb_mesh_extreme(
        search_direction, start_idx, vertices, connections,
        shortcut_connections):
    """Hill climbing to find support point of mesh.

    Parameters
    ----------
    search_direction : array, shape (3,)
        Search direction in mesh frame.

    start_idx : int
        Index of vertex from which we start hill climbing.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    connections : dict
        Maps vertex indices to neighboring vertex indices.

    shortcut_connections : array, shape (n_shortcut_connections,)
        Indices of extreme vertices of the mesh in mesh frame.
    """
    search_direction = np.ascontiguousarray(search_direction)
    best_idx = start_idx

    if shortcut_connections is not None:
        for connected_idx in shortcut_connections:
            vertex_diff = np.ascontiguousarray(
                vertices[connected_idx] - vertices[best_idx])
            projected_length = search_direction.dot(vertex_diff)
            if projected_length > 0.0:
                best_idx = connected_idx

    converged = False
    while not converged:
        converged = True
        for connected_idx in connections[best_idx]:
            vertex_diff = np.ascontiguousarray(
                vertices[connected_idx] - vertices[best_idx])
            projected_length = search_direction.dot(vertex_diff)
            if projected_length > 0.0:
                best_idx = connected_idx
                converged = False

    return best_idx


class MeshSupportFunction:
    """Standard mesh support function.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.
    """
    def __init__(self, mesh2origin, vertices, triangles):
        self.mesh2origin = mesh2origin
        self.vertices = vertices
        self.first_idx = 0

    def update_pose(self, mesh2origin):
        """Update pose.

        Parameters
        ----------
        mesh2origin : array, shape (4, 4)
            New pose of the mesh.
        """
        self.mesh2origin = mesh2origin

    def __call__(self, search_direction):
        """Support function.

        Parameters
        ----------
        search_direction : array, shape (3,)
            Search direction.

        Returns
        -------
        idx : int
            Index of support point.

        support_point : array, shape (3,)
            Support point.
        """
        search_direction_in_mesh = np.dot(
            self.mesh2origin[:3, :3].T, search_direction)
        idx = np.argmax(self.vertices.dot(search_direction_in_mesh))
        return idx, self.mesh2origin[:3, 3] + np.dot(
            self.mesh2origin[:3, :3], self.vertices[idx])


def make_convex_mesh(vertices):
    """Make convex mesh from vertices.

    Parameters
    ----------
    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh (not necessarily convex, but might be).

    Returns
    -------
    triangles : array, shape (n_triangles, 3)
        Indices of vertices forming the faces of a convex mesh. Normals of the
        faces point to the outside of the mesh.
    """
    vertices = vertices - np.mean(vertices, axis=0)
    ch = ConvexHull(vertices)
    triangles = ch.simplices
    faces = vertices[triangles]
    A = faces[:, 2] - faces[:, 0]
    B = faces[:, 1] - faces[:, 0]
    face_normals = np.cross(A, B)
    face_centers = np.mean(faces, axis=1)
    angles = angles_between_vectors(face_normals, face_centers)
    indices = np.where(angles < 0.5 * np.pi)[0]
    triangles[indices] = triangles[indices, ::-1]
    return triangles



import numba
import numpy as np


class Simplex:
    def __init__(self):
        self.v = np.empty((4, 3))  # Vertices of the Minkowski difference
        self.v1 = np.empty((4, 3))  # Vertices of collider 1
        self.v2 = np.empty((4, 3))  # Vertices of collider 2
        self.n_points = 0

    def __len__(self):
        return self.n_points

    def add_point(self, v, v1, v2):
        self.v[self.n_points] = v
        self.v1[self.n_points] = v1
        self.v2[self.n_points] = v2
        self.n_points += 1


def support_function(collider1, collider2, search_direction):
    v1 = collider1.support_function(search_direction)
    v2 = collider2.support_function(-search_direction)
    return make_support_point(v1, v2)


@numba.njit(cache=True)
def make_support_point(v1, v2):
    return v1 - v2, v1, v2


def minkowski_sum(vertices1, vertices2):
    """Minkowski sum of two sets of vertices.

    Parameters
    ----------
    vertices1 : array, shape (n_vertices1, 3)
        First set of vertices.

    vertices2 : array, shape (n_vertices2, 3)
        Second set of vertices.

    Returns
    -------
    ms : array, shape (n_vertices1 * n_vertices2, 3)
        Sums of all pairs of vertices from first and second set.
    """
    return np.array([v1 + v2 for v1 in vertices1 for v2 in vertices2])



import math
import numpy as np


EPSILON = 10.0 * np.finfo(float).eps


def gjk(collider1, collider2):
    """Gilbert-Johnson-Keerthi (GJK) algorithm for distance calculation.

    The GJK algorithm only works for convex shapes. Concave objects have to be
    decomposed into convex shapes first.

    Based on the translation to C of the original Fortran implementation:
    Ruspini, Diego. gilbert.c, a C version of the original Fortran
    implementation of the GJK algorithm.
    ftp://labrea.stanford.edu/cs/robotics/sean/distance/gilbert.c,
    also available from http://realtimecollisiondetection.net/files/gilbert.c

    The original publication describing the algorithm is:
    E.G. Gilbert, D.W. Johnson, S.S. Keerthi: A fast procedure for computing
    the distance between complex objects in three-dimensional space, IEEE
    Journal of Robotics and Automation (1988),
    https://graphics.stanford.edu/courses/cs448b-00-winter/papers/gilbert.pdf

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    Returns
    -------
    distance : float
        The shortest distance between two convex shapes.

    closest_point1 : array, shape (3,)
        Closest point on first convex shape.

    closest_point2 : array, shape (3,)
        Closest point on second convex shape.

    simplex : array, shape (4, 3)
        Simplex defined by 4 points of the Minkowski difference between
        vertices of the two colliders.
    """
    collider1 = VertexCachedCollider(collider1)
    collider2 = VertexCachedCollider(collider2)

    solution = Solution()
    simplex = SimplexInfo()
    old_simplex = SimplexInfo()

    new_idx1, new_vertex1 = collider1.first_vertex()
    new_idx2, new_vertex2 = collider2.first_vertex()
    simplex.set_first_point(new_idx1, new_idx2, new_vertex1 - new_vertex2)

    iteration = 0
    backup = False
    while True:
        iteration += 1

        new_solution, backup = distance_subalgorithm_with_backup_procedure(
            simplex, solution, backup)

        no_improvement = new_solution.distance_squared >= solution.distance_squared
        simplex_is_tetrahedron = len(simplex) == 4
        if no_improvement or simplex_is_tetrahedron:
            if backup:
                closest_point1 = compute_point(
                    collider1, solution.barycentric_coordinates[:len(simplex)],
                    simplex.indices_polytope1[:len(simplex)])
                closest_point2 = compute_point(
                    collider2, solution.barycentric_coordinates[:len(simplex)],
                    simplex.indices_polytope2[:len(simplex)])

                if simplex_is_tetrahedron:
                    # Make sure intersection has zero distance
                    closest_point1[:] = 0.5 * (closest_point1 + closest_point2)
                    closest_point2[:] = closest_point1
                    distance = 0.0
                else:
                    distance = new_solution.distance

                return distance, closest_point1, closest_point2, simplex.points
            else:
                backup = True
                if iteration != 1:
                    simplex.copy_from(old_simplex)
        else:
            solution = new_solution

            _find_new_supporting_point(collider1, collider2, simplex, solution)

            old_simplex.copy_from(simplex)
            if len(simplex) == 4:
                simplex.reorder(simplex.nondecreasing_ordered_indices())


gjk_with_simplex = gjk  # for backward compatibility


class VertexCachedCollider:
    """Caches vertices from a collider.

    Many colliders do not store their vertices explicitly. However, GJK only
    stores indices to the vertices. Hence, we will cache requested vertices
    of a collider here.

    Parameters
    ----------
    collider : ConvexCollider
        Collider object.
    """
    def __init__(self, collider):
        self.collider = collider
        self.vertices_ = []

    def first_vertex(self):
        """Get vertex from collider to initialize GJK algorithm.

        Returns
        -------
        vertex_idx : int
            Index in vertex cache.

        vertex : array, shape (3,)
            Vertex from collider.
        """
        vertex = self.collider.first_vertex()
        vertex_idx = len(self.vertices_)
        self.vertices_.append(vertex)
        return vertex_idx, vertex

    def support_function(self, search_direction):
        """Support function for collider.

        Parameters
        ----------
        search_direction : array, shape (3,)
            Direction in which we search for support point of the collider.

        Returns
        -------
        vertex_idx : int
            Index in vertex cache.

        support_point : array, shape (3,)
            Extreme point along search direction.
        """
        vertex = self.collider.support_function(search_direction)
        vertex_idx = len(self.vertices_)
        self.vertices_.append(vertex)
        return vertex_idx, vertex


def _find_new_supporting_point(collider1, collider2, simplex, solution):
    new_index1, new_vertex1 = collider1.support_function(-solution.search_direction)
    new_index2, new_vertex2 = collider2.support_function(solution.search_direction)
    new_simplex_point = new_vertex1 - new_vertex2
    simplex.add_new_point(new_index1, new_index2, new_simplex_point)


def compute_point(vertex_cache, barycentric_coordinates, indices):
    """Compute point from barycentric coordinates.

    Parameters
    ----------
    vertex_cache : VertexCachedCollider
        Collider that contains cached vertices.

    barycentric_coordinates : array, shape (n_vertices,)
        Barycentric coordinates of the point that we compute.

    indices : array, shape (n_vertices,)
        Vertex indices to which the barycentric coordinates apply.

    Returns
    -------
    point : array, shape (3,)
        Point that we compute from barycentric coordinates.
    """
    return np.dot(barycentric_coordinates,
                  [vertex_cache.vertices_[i] for i in indices])


class Solution:
    """Represents current best solution and search direction.

    Attributes
    ----------
    search_direction : array, shape (3,)
        Near point to the convex hull of the points in simplex.

    barycentric_coordinates : array, shape (n_simplex_points,)
        The barycentric coordinates of search_direction, i.e.,
        search_direction = barycentric_coordinates[0]*simplex[1] + ...
        + barycentric_coordinates(n_simplex_points)*simplex[n_simplex_points-1],
        barycentric_coordinates[k] > 0.0 for k=0,...,n_simplex_points-1, and,
        barycentric_coordinates[0] + ...
        + barycentric_coordinates[n_simplex_points-1] = 1.0.

    distance_squared : float
        Squared distance to origin.
    """
    def __init__(self):
        self.barycentric_coordinates = np.empty(4, dtype=float)
        self.barycentric_coordinates[0] = 1.0
        self.search_direction = np.empty(3, dtype=float)
        self.distance_squared = np.inf

    @property
    def distance(self):
        return math.sqrt(self.distance_squared)

    def from_vertex(self, simplex, vi):
        self.barycentric_coordinates[0] = 1.0
        self.search_direction = simplex.points[vi]
        self.distance_squared = simplex.dot_product_table[vi, vi]

    def from_line_segment(self, simplex, vi, a, b):
        coords_sum = a + b
        self.barycentric_coordinates[0] = a / coords_sum
        self.barycentric_coordinates[1] = 1.0 - self.barycentric_coordinates[0]
        self.search_direction = self.barycentric_coordinates[:2].dot(simplex.points[vi])
        self.distance_squared = np.dot(self.search_direction, self.search_direction)

    def from_face(self, simplex, vi, a, b, c):
        coords_sum = a + b + c
        self.barycentric_coordinates[0] = a / coords_sum
        self.barycentric_coordinates[1] = b / coords_sum
        # more stable floating point operation than c / coords_sum:
        self.barycentric_coordinates[2] = 1.0 - (
            self.barycentric_coordinates[0] + self.barycentric_coordinates[1])
        self.search_direction = self.barycentric_coordinates[:3].dot(simplex.points[vi])
        self.distance_squared = np.dot(self.search_direction, self.search_direction)

    def from_tetrahedron(self, simplex, barycentric_coordinates):
        self.barycentric_coordinates[:] = barycentric_coordinates / sum(barycentric_coordinates)
        self.search_direction = self.barycentric_coordinates.dot(simplex.points)
        self.distance_squared = np.dot(self.search_direction, self.search_direction)

    def copy_from(self, solution, n_simplex_points):
        self.barycentric_coordinates[:n_simplex_points] = \
            solution.barycentric_coordinates[:n_simplex_points]
        self.search_direction = solution.search_direction
        self.distance_squared = solution.distance_squared


class SimplexInfo:
    """Simplex and additional data.

    Attributes
    ----------
    n_simplex_points : int
        Number of current simplex points.

    points : array, shape (n_simplex_points, 3)
      Current simplex.

    dot_product_table : array, shape (n_simplex_points, n_simplex_points)
        dot_product_table[i, j] = Inner product of simplex[i] and simplex[j].
        Note that only elements i >= j are used.

    indices_polytope1 : array, shape (n_simplex_points,)
        Index vector for first polytope. For k = 1, ..., n_simplex_points,
        simplex[k] = vertices1[indices_polytope1[k]]
        - vertices2[indices_polytope2[k]].

    indices_polytope2 : array, shape (n_simplex_points,)
        Index vectors for first and second polytope. For k = 1, ...,
        n_simplex_points, simplex[k] = vertices1[indices_polytope1[k]]
        - vertices2[indices_polytope2[k]].
    """
    def __init__(self):
        self.n_simplex_points = 0
        self.points = np.empty((4, 3), dtype=float)
        self.dot_product_table = np.empty((4, 4), dtype=float)
        self.indices_polytope1 = np.empty(4, dtype=int)
        self.indices_polytope2 = np.empty(4, dtype=int)

    def set_first_point(self, new_index1, new_index2, new_simplex_point):
        self.indices_polytope1[0] = new_index1
        self.indices_polytope2[0] = new_index2
        self.points[0] = new_simplex_point
        self.n_simplex_points += 1
        self.dot_product_table[:self.n_simplex_points, 0] = np.dot(
            self.points[:self.n_simplex_points], self.points[0])

    def copy_from(self, simplex):
        self.n_simplex_points = len(simplex)
        self.points[:len(simplex)] = simplex.points[:len(simplex)]
        self.indices_polytope1[:len(simplex)] = simplex.indices_polytope1[:len(simplex)]
        self.indices_polytope2[:len(simplex)] = simplex.indices_polytope2[:len(simplex)]
        self.dot_product_table[:self.n_simplex_points, :self.n_simplex_points] = simplex.dot_product_table[
            :self.n_simplex_points, :self.n_simplex_points]

    def reorder(self, ordered_indices):
        self.n_simplex_points = len(ordered_indices)
        self.indices_polytope1[:self.n_simplex_points] = self.indices_polytope1[ordered_indices]
        self.indices_polytope2[:self.n_simplex_points] = self.indices_polytope2[ordered_indices]
        self.points[:self.n_simplex_points] = self.points[ordered_indices]
        self.dot_product_table = self.points.dot(self.points.T)

    def add_new_point(self, new_index1, new_index2, new_simplex_point):
        self._move_first_point_to_last_spot()
        self.set_first_point(new_index1, new_index2, new_simplex_point)

    def _move_first_point_to_last_spot(self):
        self.indices_polytope1[self.n_simplex_points] = self.indices_polytope1[0]
        self.indices_polytope2[self.n_simplex_points] = self.indices_polytope2[0]
        self.points[self.n_simplex_points] = self.points[0]
        self.dot_product_table[self.n_simplex_points, :self.n_simplex_points] = self.dot_product_table[
                                                                                :self.n_simplex_points, 0]
        self.dot_product_table[self.n_simplex_points, self.n_simplex_points] = self.dot_product_table[0, 0]

    def _move_vertex(self, old_index, new_index):
        if old_index == new_index:
            return
        self.indices_polytope1[new_index] = self.indices_polytope1[old_index]
        self.indices_polytope2[new_index] = self.indices_polytope2[old_index]
        self.points[new_index] = self.points[old_index]

    def select_vertex(self, i):
        self.n_simplex_points = 1
        self._move_vertex(i, 0)
        self.dot_product_table[0, 0] = self.dot_product_table[i, i]

    def select_line_segment(self, i, j):
        self.n_simplex_points = 2
        if i != 0:
            self._move_vertex(i, 0)
            self.dot_product_table[0, 0] = self.dot_product_table[i, i]
            idx1, idx2 = (j, i) if i < j else (i, j)
            self.dot_product_table[1, 0] = self.dot_product_table[idx1, idx2]
        if j != 1:
            self._move_vertex(j, 1)
            idx1, idx2 = (j, i) if i < j else (i, j)
            self.dot_product_table[1, 0] = self.dot_product_table[idx1, idx2]
            self.dot_product_table[1, 1] = self.dot_product_table[j, j]

    def select_face(self, i, j, k):
        self.n_simplex_points = 3
        if i != 0:
            self._move_vertex(i, 0)
            self.dot_product_table[0, 0] = self.dot_product_table[i, i]
            idx1, idx2 = (j, i) if i < j else (i, j)
            self.dot_product_table[1, 0] = self.dot_product_table[idx1, idx2]
            idx1, idx2 = (k, i) if i < k else (i, k)
            self.dot_product_table[2, 0] = self.dot_product_table[idx1, idx2]
        if j != 1:
            self._move_vertex(j, 1)
            idx1, idx2 = (j, i) if i < j else (i, j)
            self.dot_product_table[1, 0] = self.dot_product_table[idx1, idx2]
            self.dot_product_table[1, 1] = self.dot_product_table[j, j]
            idx1, idx2 = (j, k) if k < j else (k, j)
            self.dot_product_table[2, 1] = self.dot_product_table[idx1, idx2]
        if k != 2:
            self._move_vertex(k, 2)
            idx1, idx2 = (i, k) if k < i else (k, i)
            self.dot_product_table[2, 0] = self.dot_product_table[idx1, idx2]
            idx1, idx2 = (j, k) if k < j else (k, j)
            self.dot_product_table[2, 1] = self.dot_product_table[idx1, idx2]
            self.dot_product_table[2, 2] = self.dot_product_table[k, k]

    def nondecreasing_ordered_indices(self):
        # fast version of np.hstack(((0,), 1 + np.argsort(self.dot_product_table[1:, 0])))
        ordered_indices = np.empty(4, dtype=int)
        ordered_indices[:3] = 0, 1, 2
        if self.dot_product_table[2, 0] < self.dot_product_table[1, 0]:
            ordered_indices[1] = 2
            ordered_indices[2] = 1
        ii = ordered_indices[1]
        if self.dot_product_table[3, 0] < self.dot_product_table[ii, 0]:
            ordered_indices[3] = ordered_indices[2]
            ordered_indices[2] = ordered_indices[1]
            ordered_indices[1] = 3
        else:
            ii = ordered_indices[2]
            if self.dot_product_table[3, 0] < self.dot_product_table[ii, 0]:
                ordered_indices[3] = ordered_indices[2]
                ordered_indices[2] = 3
            else:
                ordered_indices[3] = 3
        return ordered_indices

    def __len__(self):
        return self.n_simplex_points


def distance_subalgorithm_with_backup_procedure(simplex, solution, backup=False):
    """Johnson's distance subalgorithm.

    Implements, in a very efficient way, the distance subalgorithm of finding
    the point of minimum norm in the convex hull of four or less points
    in 3D space. The procedure and its efficient FORTRAN implementation
    are both due to D.W. Johnson. Although this subroutine is quite long,
    only a very small part of it will be executed on each call. Refer to
    sections 5 and 6 of the report mentioned in routine DIST3 for details
    concerning the distance subalgorithm. Converted to C be Diego C. Ruspini
    3/25/93.

    This function also determines an affinely independent subset of the
    points such that search_direction is a near point to the affine hull of
    the points in the subset. The variable simplex is modified so that, on
    output, it corresponds to this subset of points.

    Parameters
    ----------
    simplex : array, shape (n_simplex_points, 3)
      Current simplex.

    solution : Solution
        Represents current best solution and search direction.

    backup : bool, optional (default: False)
        Perform backup procedure.

    Returns
    -------
    solution : Solution
        Current solution.

    backup : int
        Perform backup procedure.
    """
    d = BarycentricCoordinates()

    if backup:
        return backup_procedure(simplex, solution, d, backup), backup
    else:
        try:
            return distance_subalgorithm(simplex, d), backup
        except np.linalg.LinAlgError:
            backup = True
            return backup_procedure(simplex, solution, d, backup), backup


class BarycentricCoordinates:
    """Barycentric coordinates.

    Attributes
    ----------
    d : array, shape (4, 15)
        All barycentric coordinates of all vertices in all 15 cases. Each row
        corresponds to one vertex. The column index identifies one of the 15
        cases:
        * 0: vertex 0
        * 1: vertex 1
        * 2: line segment 0-1
        * 3: vertex 3
        * 4: line segment 0-2
        * 5: line segment 1-2
        * 6: face 0-1-2
        * 7: vertex 4
        * 8: line segment 0-3
        * 9: line segment 1-3
        * 10: line segment 2-3
        * 11: face 0-1-3
        * 12: face 0-2-3
        * 13: face 1-2-3
        * 14: tetrahedron 0-1-2-3
    """
    def __init__(self):
        self.d = np.empty((4, 15), dtype=float)
        self.d[0, 0] = 1.0
        self.d[1, 1] = 1.0
        self.d[2, 3] = 1.0
        self.d[3, 7] = 1.0

    def line_segment_coordinates_0(self, simplex):
        self.d[1, 2] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[1, 0]

    def line_segment_coordinates_1(self, simplex):
        self.d[0, 2] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[1, 0]

    def face_coordinates_0(self, simplex):
        self.line_segment_coordinates_0(simplex)
        self.d[2, 4] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[2, 0]

    def face_coordinates_1(self, simplex):
        e132 = simplex.dot_product_table[1, 0] - simplex.dot_product_table[2, 1]
        self.line_segment_coordinates_1(simplex)
        self.d[2, 6] = self.d[0, 2] * self.d[2, 4] + self.d[1, 2] * e132

    def face_coordinates_2(self, simplex):
        e123 = simplex.dot_product_table[2, 0] - simplex.dot_product_table[2, 1]
        self.d[0, 4] = simplex.dot_product_table[2, 2] - simplex.dot_product_table[2, 0]
        self.d[1, 6] = self.d[0, 4] * self.d[1, 2] + self.d[2, 4] * e123
        return e123

    def face_coordinates_3(self, simplex, e123):
        e213 = -e123
        self.d[1, 5] = simplex.dot_product_table[2, 2] - simplex.dot_product_table[2, 1]
        self.d[2, 5] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[2, 1]
        self.d[0, 6] = self.d[1, 5] * self.d[0, 2] + self.d[2, 5] * e213
        return e213

    def tetrahedron_coordinates_0(self, simplex):
        self.face_coordinates_0(simplex)
        self.d[3, 8] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[3, 0]

    def tetrahedron_coordinates_1(self, simplex):
        e132 = simplex.dot_product_table[1, 0] - simplex.dot_product_table[2, 1]
        e142 = simplex.dot_product_table[1, 0] - simplex.dot_product_table[3, 1]
        self.d[0, 2] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[1, 0]
        self.d[2, 6] = self.d[0, 2] * self.d[2, 4] + self.d[1, 2] * e132
        self.d[3, 11] = self.d[0, 2] * self.d[3, 8] + self.d[1, 2] * e142
        return e132, e142

    def tetrahedron_coordinates_2(self, simplex):
        e123 = simplex.dot_product_table[2, 0] - simplex.dot_product_table[2, 1]
        e143 = simplex.dot_product_table[2, 0] - simplex.dot_product_table[3, 2]
        self.d[0, 4] = simplex.dot_product_table[2, 2] - simplex.dot_product_table[2, 0]
        self.d[1, 6] = self.d[0, 4] * self.d[1, 2] + self.d[2, 4] * e123
        self.d[3, 12] = self.d[0, 4] * self.d[3, 8] + self.d[2, 4] * e143
        return e123, e143

    def tetrahedron_coordinates_3(self, simplex, e123, e142, e143):
        self.d[1, 5] = simplex.dot_product_table[2, 2] - simplex.dot_product_table[2, 1]
        self.d[2, 5] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[2, 1]
        e213 = -e123
        self.d[0, 6] = self.d[1, 5] * self.d[0, 2] + self.d[2, 5] * e213
        self.d[3, 14] = self.d[0, 6] * self.d[3, 8] + self.d[1, 6] * e142 + self.d[2, 6] * e143
        return e213

    def tetrahedron_coordinates_4(self, simplex):
        e124 = simplex.dot_product_table[3, 0] - simplex.dot_product_table[3, 1]
        e134 = simplex.dot_product_table[3, 0] - simplex.dot_product_table[3, 2]
        self.d[0, 8] = simplex.dot_product_table[3, 3] - simplex.dot_product_table[3, 0]
        self.d[1, 11] = self.d[0, 8] * self.d[1, 2] + self.d[3, 8] * e124
        self.d[2, 12] = self.d[0, 8] * self.d[2, 4] + self.d[3, 8] * e134
        return e124, e134

    def tetrahedron_coordinates_5(self, simplex, e124, e132, e134):
        self.d[1, 9] = simplex.dot_product_table[3, 3] - simplex.dot_product_table[3, 1]
        self.d[3, 9] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[3, 1]
        e214 = -e124
        self.d[0, 11] = self.d[1, 9] * self.d[0, 2] + self.d[3, 9] * e214
        self.d[2, 14] = self.d[0, 11] * self.d[2, 4] + self.d[1, 11] * e132 + self.d[3, 11] * e134
        return e214

    def tetrahedron_coordinates_6(self, simplex, e123, e124, e134):
        self.d[2, 10] = simplex.dot_product_table[3, 3] - simplex.dot_product_table[3, 2]
        self.d[3, 10] = simplex.dot_product_table[2, 2] - simplex.dot_product_table[3, 2]
        e314 = -e134
        self.d[0, 12] = self.d[2, 10] * self.d[0, 4] + self.d[3, 10] * e314
        self.d[1, 14] = self.d[0, 12] * self.d[1, 2] + self.d[2, 12] * e123 + self.d[3, 12] * e124

    def tetrahedron_coordinates_7(self, simplex, e213, e214):
        e243 = simplex.dot_product_table[2, 1] - simplex.dot_product_table[3, 2]
        self.d[3, 13] = self.d[1, 5] * self.d[3, 9] + self.d[2, 5] * e243
        e234 = simplex.dot_product_table[3, 1] - simplex.dot_product_table[3, 2]
        self.d[2, 13] = self.d[1, 9] * self.d[2, 5] + self.d[3, 9] * e234
        e324 = -e234
        self.d[1, 13] = self.d[2, 10] * self.d[1, 5] + self.d[3, 10] * e324
        self.d[0, 14] = self.d[1, 13] * self.d[0, 2] + self.d[2, 13] * e213 + self.d[3, 13] * e214

    def backup_line_segments(self, simplex):
        self.d[1, 2] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[1, 0]
        self.d[0, 2] = simplex.dot_product_table[1, 1] - simplex.dot_product_table[1, 0]

    def backup_faces(self, simplex):
        self.backup_line_segments(simplex)
        self.d[2, 4] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[2, 0]
        e132 = simplex.dot_product_table[1, 0] - simplex.dot_product_table[2, 1]
        self.d[2, 6] = self.d[0, 2] * self.d[2, 4] + self.d[1, 2] * e132
        e123 = self.face_coordinates_2(simplex)
        e213 = self.face_coordinates_3(simplex, e123)
        return e132, e123, e213

    def backup_tetrahedron(self, simplex):
        e132, e123, e213 = self.backup_faces(simplex)
        self.d[3, 8] = simplex.dot_product_table[0, 0] - simplex.dot_product_table[3, 0]
        e142 = simplex.dot_product_table[1, 0] - simplex.dot_product_table[3, 1]
        self.d[3, 11] = self.d[0, 2] * self.d[3, 8] + self.d[1, 2] * e142
        e143 = simplex.dot_product_table[2, 0] - simplex.dot_product_table[3, 2]
        self.d[3, 12] = self.d[0, 4] * self.d[3, 8] + self.d[2, 4] * e143
        self.d[3, 14] = self.d[0, 6] * self.d[3, 8] + self.d[1, 6] * e142 + self.d[2, 6] * e143
        e124, e134 = self.tetrahedron_coordinates_4(simplex)
        e214 = self.tetrahedron_coordinates_5(simplex, e124, e132, e134)
        self.tetrahedron_coordinates_6(simplex, e123, e124, e134)
        self.tetrahedron_coordinates_7(simplex, e213, e214)

    def vertex_0_of_line_segment_optimal(self):
        return self.d[1, 2] <= 0.0

    def line_segment_01_of_line_segment_optimal(self):
        return not (self.vertex_1_of_line_segment_optimal() or self.vertex_0_of_line_segment_optimal())

    def vertex_1_of_line_segment_optimal(self):
        return self.d[0, 2] <= 0.0

    def vertex_0_of_face_optimal(self):
        return not (self.d[1, 2] > 0.0 or self.d[2, 4] > 0.0)

    def line_segment_01_of_face_optimal(self):
        return self.line_segment_01_of_line_segment_optimal() and not self.d[2, 6] > 0.0

    def line_segment_02_of_face_optimal(self):
        return not (self.d[0, 4] <= 0.0 or self.d[1, 6] > 0.0 or self.d[2, 4] <= 0.0)

    def face_012_of_face_optimal(self):
        return not (self.d[0, 6] <= 0.0 or self.d[1, 6] <= 0.0 or self.d[2, 6] <= 0.0)

    def vertex_1_of_face_optimal(self):
        return not (self.d[0, 2] > 0.0 or self.d[2, 5] > 0.0)

    def vertex_2_of_face_optimal(self):
        return not (self.d[0, 4] > 0.0 or self.d[1, 5] > 0.0)

    def line_segment_12_of_face_optimal(self):
        return not self.d[0, 6] > 0.0 and self.check_line_segment_12_of_face()

    def vertex_0_of_tetrahedron_optimal(self):
        return self.vertex_0_of_face_optimal() and not self.d[3, 8] > 0.0

    def line_segment_01_of_tetrahedron_optimal(self):
        return self.line_segment_01_of_face_optimal() and not self.d[3, 11] > 0.0

    def line_segment_02_of_tetrahedron_optimal(self):
        return self.line_segment_02_of_face_optimal() and not self.d[3, 12] > 0.0

    def face_012_of_tetrahedron_optimal(self):
        return self.face_012_of_face_optimal() and not self.d[3, 14] > EPSILON

    def line_segment_03_of_tetrahedron_optimal(self):
        return not (self.d[1, 11] > 0.0 or self.d[2, 12] > 0.0) and self.check_line_segment_03_of_tetrahedron()

    def face_013_of_tetrahedron_optimal(self):
        return not self.d[2, 14] > EPSILON and self.check_face_013_of_tetrahedron()

    def face_023_of_tetrahedron_optimal(self):
        return not self.d[1, 14] > EPSILON and self.check_face_023_of_tetrahedron()

    def convex_hull_of_tetrahedron_optimal(self):
        return not (self.d[0, 14] <= EPSILON or self.d[1, 14] <= EPSILON
                    or self.d[2, 14] <= EPSILON or self.d[3, 14] <= EPSILON)

    def vertex_1_of_tetrahedron_optimal(self):
        return self.vertex_1_of_face_optimal() and not self.d[3, 9] > 0.0

    def vertex_2_of_tetrahedron_optimal(self):
        return self.vertex_2_of_face_optimal() and not self.d[3, 10] > 0.0

    def vertex_3_of_tetrahedron_optimal(self):
        return not (self.d[0, 8] > 0.0 or self.d[1, 9] > 0.0 or self.d[2, 10] > 0.0)

    def line_segment_12_of_tetrahedron_optimal(self):
        return self.line_segment_12_of_face_optimal() and not self.d[3, 13] > 0.0

    def line_segment_13_of_tetrahedron_optimal(self):
        return not (self.d[0, 11] > 0.0 or self.d[2, 13] > 0.0) and self.check_line_segment_13_of_tetrahedron()

    def line_segment_23_of_tetrahedron_optimal(self):
        return not (self.d[0, 12] > 0.0 or self.d[1, 13] > 0.0) and self.check_line_segment_23_of_tetrahedron()

    def face_123_of_tetrahedron_optimal(self):
        return not self.d[0, 14] > EPSILON and self.check_face_123_of_tetrahedron()

    def check_line_segment_02_of_face(self):
        return not (self.d[0, 4] <= 0.0 or self.d[2, 4] <= 0.0)

    def check_face_012_of_face(self):
        return not (self.d[0, 6] <= 0.0 or self.d[1, 6] <= 0.0 or self.d[2, 6] <= 0.0)

    def check_line_segment_12_of_face(self):
        return not (self.d[1, 5] <= 0.0 or self.d[2, 5] <= 0.0)

    def check_line_segment_03_of_tetrahedron(self):
        return not (self.d[0, 8] <= 0.0 or self.d[3, 8] <= 0.0)

    def check_face_013_of_tetrahedron(self):
        return not (self.d[0, 11] <= 0.0 or self.d[1, 11] <= 0.0 or self.d[3, 11] <= 0.0)

    def check_face_023_of_tetrahedron(self):
        return not (self.d[0, 12] <= 0.0 or self.d[2, 12] <= 0.0 or self.d[3, 12] <= 0.0)

    def check_line_segment_13_of_tetrahedron(self):
        return not (self.d[1, 9] <= 0.0 or self.d[3, 9] <= 0.0)

    def check_line_segment_23_of_tetrahedron(self):
        return not (self.d[2, 10] <= 0.0 or self.d[3, 10] <= 0.0)

    def check_face_123_of_tetrahedron(self):
        return not (self.d[1, 13] <= 0.0 or self.d[2, 13] <= 0.0 or self.d[3, 13] <= 0.0)


def distance_subalgorithm(simplex, d):
    """Johnson's distance subalgorithm."""
    if len(simplex) == 1:
        solution = Solution()
        solution.from_vertex(simplex, 0)
        return solution
    elif len(simplex) == 2:
        return _distance_subalgorithm_line_segment(simplex, d)
    elif len(simplex) == 3:
        return _distance_subalgorithm_face(simplex, d)
    else:
        assert len(simplex) == 4
        return _distance_subalgorithm_tetrahedron(simplex, d)


def _distance_subalgorithm_line_segment(simplex, d):
    solution = Solution()
    d.line_segment_coordinates_0(simplex)
    if d.vertex_0_of_line_segment_optimal():
        simplex.select_vertex(0)
        solution.from_vertex(simplex, 0)
        return solution
    d.line_segment_coordinates_1(simplex)
    if d.line_segment_01_of_line_segment_optimal():
        solution.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        return solution
    else:
        assert d.vertex_1_of_line_segment_optimal()
        simplex.select_vertex(1)
        solution.from_vertex(simplex, 0)
        return solution


def _distance_subalgorithm_face(simplex, d):
    solution = Solution()
    d.face_coordinates_0(simplex)
    if d.vertex_0_of_face_optimal():
        simplex.select_vertex(0)
        solution.from_vertex(simplex, 0)
        return solution
    d.face_coordinates_1(simplex)
    if d.line_segment_01_of_face_optimal():
        simplex.select_line_segment(0, 1)
        solution.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        return solution
    e123 = d.face_coordinates_2(simplex)
    if d.line_segment_02_of_face_optimal():
        simplex.select_line_segment(0, 2)
        solution.from_line_segment(simplex, [0, 1], d.d[0, 4], d.d[2, 4])
        return solution
    d.face_coordinates_3(simplex, e123)
    if d.face_012_of_face_optimal():
        solution.from_face(simplex, [0, 1, 2], d.d[0, 6], d.d[1, 6], d.d[2, 6])
        return solution
    if d.vertex_1_of_face_optimal():
        simplex.select_vertex(1)
        solution.from_vertex(simplex, 0)
        return solution
    if d.vertex_2_of_face_optimal():
        simplex.select_vertex(2)
        solution.from_vertex(simplex, 0)
        return solution
    if d.line_segment_12_of_face_optimal():
        simplex.select_line_segment(2, 1)
        solution.from_line_segment(simplex, [1, 0], d.d[1, 5], d.d[2, 5])
        return solution
    raise np.linalg.LinAlgError("Numerical problem, backup procedure required")


def _distance_subalgorithm_tetrahedron(simplex, d):
    solution = Solution()
    d.tetrahedron_coordinates_0(simplex)
    if d.vertex_0_of_tetrahedron_optimal():
        simplex.select_vertex(0)
        solution.from_vertex(simplex, 0)
        return solution
    e132, e142 = d.tetrahedron_coordinates_1(simplex)
    if d.line_segment_01_of_tetrahedron_optimal():
        simplex.select_line_segment(0, 1)
        solution.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        return solution
    e123, e143 = d.tetrahedron_coordinates_2(simplex)
    if d.line_segment_02_of_tetrahedron_optimal():
        simplex.select_line_segment(0, 2)
        solution.from_line_segment(simplex, [0, 1], d.d[0, 4], d.d[2, 4])
        return solution
    e213 = d.tetrahedron_coordinates_3(simplex, e123, e142, e143)
    if d.face_012_of_tetrahedron_optimal():
        simplex.select_face(0, 1, 2)
        solution.from_face(simplex, [0, 1, 2], d.d[0, 6], d.d[1, 6], d.d[2, 6])
        return solution
    e124, e134 = d.tetrahedron_coordinates_4(simplex)
    if d.line_segment_03_of_tetrahedron_optimal():
        simplex.select_line_segment(0, 3)
        solution.from_line_segment(simplex, [0, 1], d.d[0, 8], d.d[3, 8])
        return solution
    e214 = d.tetrahedron_coordinates_5(simplex, e124, e132, e134)
    if d.face_013_of_tetrahedron_optimal():
        simplex.select_face(0, 1, 3)
        solution.from_face(simplex, [0, 1, 2], d.d[0, 11], d.d[1, 11], d.d[3, 11])
        return solution
    d.tetrahedron_coordinates_6(simplex, e123, e124, e134)
    if d.face_023_of_tetrahedron_optimal():
        simplex.select_face(0, 3, 2)
        solution.from_face(simplex, [0, 1, 2], d.d[0, 12], d.d[3, 12], d.d[2, 12])
        return solution
    d.tetrahedron_coordinates_7(simplex, e213, e214)
    if d.convex_hull_of_tetrahedron_optimal():
        solution.from_tetrahedron(simplex, d.d[:, 14])
        return solution
    if d.vertex_1_of_tetrahedron_optimal():
        simplex.select_vertex(1)
        solution.from_vertex(simplex, 0)
        return solution
    if d.vertex_2_of_tetrahedron_optimal():
        simplex.select_vertex(2)
        solution.from_vertex(simplex, 0)
        return solution
    if d.vertex_3_of_tetrahedron_optimal():
        simplex.select_vertex(3)
        solution.from_vertex(simplex, 0)
        return solution
    if d.line_segment_12_of_tetrahedron_optimal():
        simplex.select_line_segment(2, 1)
        solution.from_line_segment(simplex, [1, 0], d.d[1, 5], d.d[2, 5])
        return solution
    if d.line_segment_13_of_tetrahedron_optimal():
        simplex.select_line_segment(3, 1)
        solution.from_line_segment(simplex, [1, 0], d.d[3, 9], d.d[1, 9])
        return solution
    if d.line_segment_23_of_tetrahedron_optimal():
        simplex.select_line_segment(2, 3)
        solution.from_line_segment(simplex, [0, 1], d.d[2, 10], d.d[3, 10])
        return solution
    if d.face_123_of_tetrahedron_optimal():
        simplex.select_face(3, 1, 2)
        solution.from_face(simplex, [1, 2, 0], d.d[2, 13], d.d[3, 13], d.d[1, 13])
        return solution
    raise np.linalg.LinAlgError("Numerical problem, backup procedure required")


def backup_procedure(simplex, solution, d, backup=True):
    """Backup procedure.

    Johnson's distance subalgorithm is affected by rounding errors in floating
    point operations in cases of degenerate simplices. This backup procedure
    addresses the problem. It will always succeed, but is computationally more
    expensive.
    """
    if len(simplex) == 1:
        solution.from_vertex(simplex, 0)
        return solution
    elif len(simplex) == 2:
        ordered_indices = _backup_procedure_line_segment(
            simplex, backup, d, solution)
    elif len(simplex) == 3:
        ordered_indices = _backup_procedure_face(
            simplex, backup, d, solution)
    else:
        assert len(simplex) == 4
        ordered_indices = _backup_procedure_tetrahedron(
            simplex, backup, d, solution)

    simplex.reorder(ordered_indices)
    return solution


def _backup_procedure_line_segment(simplex, backup, d, solution):
    if backup:
        d.backup_line_segments(simplex)
    ordered_indices = np.empty(2, dtype=int)
    # check vertex 1
    solution.from_vertex(simplex, 0)
    n_simplex_points = 1
    ordered_indices[0] = 0
    if d.line_segment_01_of_line_segment_optimal():
        solution_d = Solution()
        solution_d.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 1
    check_vertex_2 = simplex.dot_product_table[1, 1] < solution.distance_squared
    if check_vertex_2:
        n_simplex_points = 1
        solution.from_vertex(simplex, 1)
        ordered_indices[0] = 1
    return ordered_indices[:n_simplex_points]


def _backup_procedure_face(simplex, backup, d, solution):
    if backup:
        d.backup_faces(simplex)
    ordered_indices = np.empty(3, dtype=int)
    # check vertex 1
    n_simplex_points = 1
    solution.from_vertex(simplex, 0)
    ordered_indices[0] = 0
    solution_d = Solution()
    if d.line_segment_01_of_line_segment_optimal():
        solution_d.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 1
    if d.check_line_segment_02_of_face():
        solution_d.from_line_segment(simplex, [0, 2], d.d[0, 4], d.d[2, 4])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 2
    if d.check_face_012_of_face():
        solution_d.from_face(simplex, [0, 1, 2], d.d[0, 6], d.d[1, 6], d.d[2, 6])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 3
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:3] = 0, 1, 2
    check_vertex_2 = simplex.dot_product_table[1, 1] < solution.distance_squared
    if check_vertex_2:
        n_simplex_points = 1
        solution.from_vertex(simplex, 1)
        ordered_indices[0] = 1
    check_vertex_3 = simplex.dot_product_table[2, 2] < solution.distance_squared
    if check_vertex_3:
        n_simplex_points = 1
        solution.from_vertex(simplex, 2)
        ordered_indices[0] = 2
    if d.check_line_segment_12_of_face():
        solution_d.from_line_segment(simplex, [2, 1], d.d[2, 5], d.d[1, 5])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 2, 1
    return ordered_indices[:n_simplex_points]


def _backup_procedure_tetrahedron(simplex, backup, d, solution):
    if backup:
        d.backup_tetrahedron(simplex)
    ordered_indices = np.empty(4, dtype=int)
    # check vertex 1
    n_simplex_points = 1
    solution.from_vertex(simplex, 0)
    ordered_indices[0] = 0
    solution_d = Solution()
    if d.line_segment_01_of_line_segment_optimal():
        solution_d.from_line_segment(simplex, [0, 1], d.d[0, 2], d.d[1, 2])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 1
    if d.check_line_segment_02_of_face():
        solution_d.from_line_segment(simplex, [0, 2], d.d[0, 4], d.d[2, 4])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 2
    if d.check_face_012_of_face():
        solution_d.from_face(simplex, [0, 1, 2], d.d[0, 6], d.d[1, 6], d.d[2, 6])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 3
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:3] = 0, 1, 2
    if d.check_line_segment_03_of_tetrahedron():
        solution_d.from_line_segment(simplex, [0, 3], d.d[0, 8], d.d[3, 8])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 0, 3
    if d.check_face_013_of_tetrahedron():
        solution_d.from_face(simplex, [0, 1, 3], d.d[0, 11], d.d[1, 11], d.d[3, 11])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 3
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:3] = 0, 1, 3
    if d.check_face_023_of_tetrahedron():
        solution_d.from_face(simplex, [0, 3, 2], d.d[0, 12], d.d[3, 12], d.d[2, 12])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 3
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:3] = 0, 3, 2
    if d.convex_hull_of_tetrahedron_optimal():
        solution_d.from_tetrahedron(simplex, d.d[:, 14])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 4
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:] = 0, 1, 2, 3
    check_vertex_2 = simplex.dot_product_table[1, 1] < solution.distance_squared
    if check_vertex_2:
        n_simplex_points = 1
        solution.from_vertex(simplex, 1)
        ordered_indices[0] = 1
    check_vertex_3 = simplex.dot_product_table[2, 2] < solution.distance_squared
    if check_vertex_3:
        n_simplex_points = 1
        solution.from_vertex(simplex, 2)
        ordered_indices[0] = 2
    check_vertex_4 = simplex.dot_product_table[3, 3] < solution.distance_squared
    if check_vertex_4:
        n_simplex_points = 1
        solution.from_vertex(simplex, 3)
        ordered_indices[0] = 3
    if d.check_line_segment_12_of_face():
        solution_d.from_line_segment(simplex, [2, 1], d.d[2, 5], d.d[1, 5])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 2, 1
    if d.check_line_segment_13_of_tetrahedron():
        solution_d.from_line_segment(simplex, [3, 1], d.d[3, 9], d.d[1, 9])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 3, 1
    if d.check_line_segment_23_of_tetrahedron():
        solution_d.from_line_segment(simplex, [2, 3], d.d[2, 10], d.d[3, 10])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 2
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:2] = 2, 3
    if d.check_face_123_of_tetrahedron():
        solution_d.from_face(simplex, [3, 1, 2], d.d[3, 13], d.d[1, 13], d.d[2, 13])
        if solution_d.distance_squared < solution.distance_squared:
            n_simplex_points = 3
            solution.copy_from(solution_d, n_simplex_points)
            ordered_indices[:3] = 3, 1, 2
    return ordered_indices[:n_simplex_points]



"""Visualization tools."""
import numpy as np
import pytransform3d.visualizer as pv
import open3d as o3d


class Mesh(pv.Artist):
    """A mesh.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.

    c : array-like, shape (3,), optional (default: None)
        Color(s)
    """
    def __init__(self, mesh2origin, vertices, triangles, c=None):
        self.mesh = o3d.geometry.TriangleMesh()
        self.mesh.vertices = o3d.utility.Vector3dVector(vertices)
        self.mesh.triangles = o3d.utility.Vector3iVector(triangles)
        self.mesh.compute_vertex_normals()
        if c is not None:
            self.mesh.paint_uniform_color(c)
        self.mesh2origin = None
        self.set_data(mesh2origin)

    def set_data(self, mesh2origin):
        """Update data.

        Parameters
        ----------
        mesh2origin : array-like, shape (4, 4)
            Pose of the mesh.
        """
        previous_mesh2origin = self.mesh2origin
        if previous_mesh2origin is None:
            previous_mesh2origin = np.eye(4)
        self.mesh2origin = mesh2origin

        self.mesh.transform(np.linalg.inv(previous_mesh2origin))
        self.mesh.transform(self.mesh2origin)

    @property
    def geometries(self):
        """Expose geometries.

        Returns
        -------
        geometries : list
            List of geometries that can be added to the visualizer.
        """
        return [self.mesh]



"""Expanding polytope algorithm (EPA) for collision resolution after GJK."""
import numpy as np
from .utils import norm_vector


EDGES_PER_FACE = 3


def epa(simplex, collider1, collider2, max_iter=64, max_loose_edges=32, max_faces=64, epsilon=1e-8):
    """Expanding Polytope Algorithm (EPA).

    Find minimum translation vector to resolve collision.

    Based on Kevin Moran's implementation:
    https://github.com/kevinmoran/GJK/blob/b38d923d268629f30b44c3cf6d4f9974bbcdb0d3/GJK.h
    (MIT License or Unlicense)

    Parameters
    ----------
    simplex : array, shape (4, 3)
        Simplex of Minkowski distances obtained by GJK.

    collider1 : Collider
        Convex collider 1.

    collider2 : Collider
        Convex collider 2.

    max_iter : int, optional (default: 64)
        Maximum number of iterations.

    max_loose_edges : int, optional (default: 32)
        Maximum number of loose edges per iteration.

    max_faces : int, optional (default: 64)
        Maximum number of faces in polytope.

    epsilon : float, optional (default: 1e-8)
        Floating point tolerance.

    Returns
    -------
    mtv : array, shape (3,)
        Minimum translation vector to be added to the second set of vertices
        or subtracted from the first set of vertices to resolve the collision.
        The norm of this vector is the penetration depth and the direction is
        the contact normal.

    faces : array, shape (n_faces, 4, 3)
        Faces in Minkowski difference space, defined by 3 points and its
        normal.

    success : bool
        EPA converged before maximum number of iterations was reached.
    """
    polytope = Polytope(simplex, max_faces, epsilon)
    loose_edges = LooseEdges(max_loose_edges, epsilon)

    for iteration in range(max_iter):
        min_dist, closest_face = polytope.find_face_closest_to_origin()

        search_direction = closest_face[3]
        new_vertex1 = collider1.support_function(search_direction)
        new_vertex2 = collider2.support_function(-search_direction)
        new_point = new_vertex1 - new_vertex2

        if np.dot(new_point, search_direction) - min_dist < epsilon:
            # Convergence: new point is not significantly further from origin.
            mtv = closest_face[3] * np.dot(new_point, search_direction)
            return mtv, polytope.get_all_faces(), True

        loose_edges.find_triangles_facing_point_and_store_loose_edges(
            polytope, new_point)
        polytope.extend_with_point(loose_edges, new_point)

    # Return most recent closest point
    mtv = closest_face[3] * np.dot(closest_face[0], closest_face[3])
    return mtv, polytope.get_all_faces(), False


class Polytope:
    """Array of faces, each with 3 vertices and a normal."""
    def __init__(self, simplex, max_faces, epsilon):
        self.max_faces = max_faces
        self.epsilon = epsilon
        self.faces = np.zeros((self.max_faces, 4, 3))
        self.n_faces = self._initialize_from_simplex(simplex)

    def _initialize_from_simplex(self, simplex):
        self.faces[0, :3] = simplex[:3]  # ABC
        self.faces[1, :3] = simplex[np.array((0, 2, 3), dtype=int)]  # ACD
        self.faces[2, :3] = simplex[np.array((0, 3, 1), dtype=int)]  # ADB
        self.faces[3, :3] = simplex[np.array((1, 3, 2), dtype=int)]  # BDC
        n_faces = 4
        for i in range(n_faces):
            self.compute_normal(i)
        return n_faces

    def compute_normal(self, face_idx):
        self.faces[face_idx, 3] = norm_vector(
            np.cross(self.faces[face_idx, 1] - self.faces[face_idx, 0],
                     self.faces[face_idx, 2] - self.faces[face_idx, 0]))

    def find_face_closest_to_origin(self):
        dists = np.sum(
            self.faces[:self.n_faces, 0] * self.faces[:self.n_faces, 3],
            axis=1)
        closest_face = np.argmin(dists)
        return dists[closest_face], self.faces[closest_face]

    def get_edge(self, face_idx, edge_idx):
        return np.vstack([self.faces[face_idx, edge_idx],
                          self.faces[face_idx, (edge_idx + 1) % 3]])

    def get_all_faces(self):
        return self.faces[:self.n_faces]

    def remove_face(self, i):
        self.faces[i] = self.faces[self.n_faces - 1]
        self.n_faces -= 1

    def triangle_faces_point(self, i, new_points):
        return np.dot(self.faces[i, 3],
                      new_points - self.faces[i, 0]) > self.epsilon

    def extend_with_point(self, loose_edges, new_point):
        for i in range(loose_edges.n_loose_edges):
            assert self.n_faces < self.max_faces
            if self.n_faces >= self.max_faces:
                break
            self.faces[self.n_faces, :2] = loose_edges.loose_edges[i]
            self.faces[self.n_faces, 2] = new_point
            self.compute_normal(self.n_faces)
            if np.linalg.norm(self.faces[self.n_faces, 3]) < 0.5:  # TODO is this the right solution?
                continue
            self.fix_ccw_normal_direction(self.n_faces)
            self.n_faces += 1

    def fix_ccw_normal_direction(self, face_idx, bias=1e-6):
        """Correct wrong normal direction to maintain CCW winding."""
        # Use bias in case dot result is only slightly < 0 (because origin is on face)
        if np.dot(self.faces[face_idx, 0], self.faces[face_idx, 3]) + bias < 0.0:
            temp = self.faces[face_idx, 0]
            self.faces[face_idx, 0] = self.faces[face_idx, 1]
            self.faces[face_idx, 1] = temp
            self.faces[face_idx, 3] = -self.faces[face_idx, 3]


class LooseEdges:
    """Keep track of edges we need to fix after removing faces."""
    def __init__(self, max_loose_edges, epsilon):
        self.max_loose_edges = max_loose_edges
        self.epsilon = epsilon
        self.loose_edges = np.zeros((self.max_loose_edges, 2, 3))
        self.n_loose_edges = 0

    def find_triangles_facing_point_and_store_loose_edges(self, faces, new_point):
        self.n_loose_edges = 0
        i = 0
        while i < faces.n_faces:
            if faces.triangle_faces_point(i, new_point):
                self.add_removed_triangles_edges_to_list(faces, i)
                faces.remove_face(i)
                i -= 1
            i += 1

    def add_removed_triangles_edges_to_list(self, faces, i):
        for j in range(EDGES_PER_FACE):
            current_edge = faces.get_edge(i, j)
            found_edge = False
            k = 0
            while k < self.n_loose_edges:
                if self.edge_already_in_list(k, current_edge):
                    # This assumes edge can only be shared by 2 triangles
                    # (which should be true). This also assumes shared edge
                    # will be reversed in the triangles (which should be true
                    # provided every triangle is wound CCW).
                    self.overwrite_edge_with_last_edge(k)
                    found_edge = True
                    # Exit loop because edge can only be shared once
                    break
                k += 1

            if not found_edge:
                success = self.add_edge_to_list(current_edge)
                if not success:
                    break

    def edge_already_in_list(self, edge_idx, current_edge):
        return (np.linalg.norm(self.loose_edges[edge_idx, 1] - current_edge[0]) < self.epsilon and
                np.linalg.norm(self.loose_edges[edge_idx, 0] - current_edge[1]) < self.epsilon)

    def add_edge_to_list(self, edge):
        if self.n_loose_edges >= self.max_loose_edges:
            return False
        self.loose_edges[self.n_loose_edges] = edge
        self.n_loose_edges += 1
        return True

    def overwrite_edge_with_last_edge(self, edge_idx):
        self.loose_edges[edge_idx] = self.loose_edges[self.n_loose_edges - 1]
        self.n_loose_edges -= 1



import numpy as np
import pytransform3d.transformations as pt
from ..geometry import (
    convert_segment_to_line, convert_rectangle_to_vertices,
    convert_box_to_face)
from ._line_to_box import _line_to_box
from ._rectangle import rectangle_to_rectangle


def point_to_box(point, box2origin, size, check=False):
    """Compute the shortest distance between point and box.

    Parameters
    ----------
    point : array, shape (3,)
        3D point.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    check : bool, optional (default: True)
        Check if transformation matrix is valid before inversion.

    Returns
    -------
    dist : float
        The shortest between point and box.

    closest_point_box : array, shape (3,)
        Closest point on box.
    """
    origin2box = pt.invert_transform(box2origin, check=check)
    point_in_box = origin2box[:3, 3] + origin2box[:3, :3].dot(point)
    half_size = 0.5 * size
    closest_point_in_box = np.clip(point_in_box, -half_size, half_size)
    closest_point = box2origin[:3, 3] + box2origin[:3, :3].dot(closest_point_in_box)
    return np.linalg.norm(point - closest_point), closest_point


def line_to_box(line_point, line_direction, box2origin, size):
    """Compute the shortest distance between line and box.

    Parameters
    ----------
    line_point : array, shape (3,)
        Point on line.

    line_direction : array, shape (3,)
        Direction of the line. This is assumed to be of unit length.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    Returns
    -------
    dist : float
        The shortest between line and box.

    closest_point_line : array, shape (3,)
        Closest point on line.

    closest_point_box : array, shape (3,)
        Closest point on box.
    """
    return _line_to_box(line_point, line_direction, box2origin, size)[:3]


def line_segment_to_box(segment_start, segment_end, box2origin, size):
    """Compute the shortest distance from line segment to box.

    Parameters
    ----------
    segment_start : array, shape (3,)
        Start point of segment.

    segment_end : array, shape (3,)
        End point of segment.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    Returns
    -------
    distance : float
        The shortest distance between line segment and box.

    closest_point_line_segment : array, shape (3,)
        Closest point on line segment.

    closest_point_box : array, shape (3,)
        Closest point on box.
    """
    segment_direction, segment_length = convert_segment_to_line(
        segment_start, segment_end)

    distance, closest_point_segment, closest_point_box, t_closest = _line_to_box(
        segment_start, segment_direction, box2origin, size)

    if t_closest < 0:
        distance, closest_point_box = point_to_box(
            segment_start, box2origin, size)
        closest_point_segment = segment_start
    elif t_closest > segment_length:
        distance, closest_point_box = point_to_box(
            segment_end, box2origin, size)
        closest_point_segment = segment_end

    return distance, closest_point_segment, closest_point_box


def rectangle_to_box(rectangle_center, rectangle_axes, rectangle_lengths,
                     box2origin, size, epsilon=1e-6):
    """Compute the shortest distance from rectangle to box.

    Parameters
    ----------
    rectangle_center : array, shape (3,)
        Center point of the rectangle.

    rectangle_axes : array, shape (2, 3)
        Each row is a vector of unit length, indicating the direction of one
        axis of the rectangle. Both vectors are orthogonal.

    rectangle_lengths : array, shape (2,)
        Lengths of the two sides of the rectangle.

    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Size of the box along its axes.

    epsilon : float, optional (default: 1e-6)
        Values smaller than epsilon are considered to be 0.

    Returns
    -------
    dist : float
        The shortest distance between rectangle and box.

    closest_point_rectangle : array, shape (3,)
        Closest point on the rectangle.

    closest_point_box : array, shape (3,)
        Closest point on the box.
    """
    overlap, result = _rectangle_points_in_box(
        rectangle_center, rectangle_axes, rectangle_lengths,
        box2origin, size, epsilon=epsilon)
    if overlap:
        return result
    return _rectangle_to_box_faces(
        rectangle_center, rectangle_axes, rectangle_lengths, box2origin, size,
        epsilon)


def _rectangle_points_in_box(
        rectangle_center, rectangle_axes, rectangle_lengths, box2origin, size,
        epsilon=1e-6):
    rectangle_points = convert_rectangle_to_vertices(
        rectangle_center, rectangle_axes, rectangle_lengths)
    for i in range(len(rectangle_points)):
        dist, closest_point_box = point_to_box(
            rectangle_points[i], box2origin, size)
        if dist <= epsilon:
            return True, (dist, rectangle_points[i], closest_point_box)
    return False, None


def _rectangle_to_box_faces(
        rectangle_center, rectangle_axes, rectangle_lengths, box2origin, size,
        epsilon):
    best_distance = np.finfo(float).max
    for sign in [-1, 1]:
        for i in range(3):
            face_center, face_axes, face_lengths = convert_box_to_face(
                box2origin, size, i, sign)
            dist, closest_point_rectangle, closest_point_face = rectangle_to_rectangle(
                rectangle_center, rectangle_axes, rectangle_lengths,
                face_center, face_axes, face_lengths, epsilon)
            if dist < best_distance:
                best_distance = dist
                best_closest_point_rectangle = closest_point_rectangle
                best_closest_point_box = closest_point_face

                if best_distance <= epsilon:
                    break
    return best_distance, best_closest_point_rectangle, best_closest_point_box



import math
from enum import Enum
import numba
import numpy as np

from distance3d.utils import EPSILON
from distance3d.minkowski import Simplex, support_function, make_support_point
from distance3d.distance import point_to_triangle


EPSILON_SQRT = math.sqrt(EPSILON)


def gjk_intersection(collider1, collider2, max_iterations=100):
    """Intersection test with Gilbert-Johnson-Keerthi (GJK) algorithm.

    This implementation of GJK is based on libccd (for details, see
    https://github.com/danfis/libccd). For the original code the copyright is
    of Daniel Fiser <danfis@danfis.cz>. It has been released under 3-clause BSD
    license.

    Parameters
    ----------
    collider1 : ConvexCollider
        Convex collider 1.

    collider2 : ConvexCollider
        Convex collider 2.

    max_iterations : int, optional (default: 100)
        Maximum number of iterations.

    Returns
    -------
    intersection : bool
        Do the two colliders intersect?
    """
    return _gjk(collider1, collider2, Simplex(), max_iterations)[0]


def _gjk(collider1, collider2, simplex, max_iterations):
    support_point = make_support_point(
        collider1.first_vertex(), collider2.first_vertex())
    simplex.add_point(*support_point)
    search_direction = -support_point[0]

    for _ in range(max_iterations):
        support_point = support_function(collider1, collider2, search_direction)
        is_origin = np.dot(support_point[0], support_point[0]) < EPSILON
        if is_origin:
            _set_point(simplex.v, simplex.v1, simplex.v2, 0, *support_point)
            simplex.n_points = 1
            return True, simplex

        is_before_origin = np.dot(support_point[0], search_direction) < -EPSILON_SQRT
        if is_before_origin:
            return False, simplex

        simplex.add_point(*support_point)

        gjk_state, search_direction, simplex.n_points = _refine_simplex(
            simplex.v, simplex.v1, simplex.v2, len(simplex))

        if gjk_state == GjkState.CONTACT:
            return True, simplex
        elif gjk_state == GjkState.NO_CONTACT:
            return False, simplex
        elif abs(np.dot(search_direction, search_direction)) < EPSILON:
            return False, simplex

    return False, simplex


class GjkState(Enum):
    NO_CONTACT = -1
    CONTINUE = 0
    CONTACT = 1


@numba.njit(cache=True)
def _refine_simplex(v, v1, v2, n_points):
    if n_points == 2:
        return _line_segment(v, v1, v2)
    elif n_points == 3:
        return _triangle(v, v1, v2)
    else:
        return _tetrahedron(v, v1, v2)


@numba.njit(cache=True)
def _set_point(v, v1, v2, idx, new_v, new_v1, new_v2):
    v[idx] = new_v
    v1[idx] = new_v1
    v2[idx] = new_v2


@numba.njit(cache=True)
def _line_segment(v, v1, v2):
    A = v[1], v1[1], v2[1]
    B = v[0]

    AB = B - A[0]
    AO = -A[0]
    dot = np.dot(AB, AO)

    tmp = np.cross(AB, AO)
    origin_on_AB_segment = abs(np.dot(tmp, tmp)) < EPSILON and dot > 0.0
    if origin_on_AB_segment:
        return GjkState.CONTACT, None, 0

    origin_is_outside_of_A = dot < EPSILON
    if origin_is_outside_of_A:
        _set_point(v, v1, v2, 0, *A)
        n_points = 1
        search_direction = AO
    else:  # origin is closer to line segment
        search_direction = _triple_cross(AB, AO, AB)
        n_points = 2

    return GjkState.CONTINUE, search_direction, n_points


@numba.njit(cache=True)
def _triangle(v, v1, v2):
    A = np.copy(v[2]), np.copy(v1[2]), np.copy(v1[2])
    B = np.copy(v[1]), np.copy(v1[1]), np.copy(v2[1])
    C = np.copy(v[0]), np.copy(v1[0]), np.copy(v1[0])

    touching_contact = abs(point_to_triangle(
        np.zeros(3), np.row_stack((A[0], B[0], C[0])))[0]) < EPSILON_SQRT
    if touching_contact:
        return GjkState.CONTACT, None, 0

    degenerated_triangle = (np.all(np.abs(A[0] - B[0]) < EPSILON)
                            or np.all(np.abs(A[0] - C[0]) < EPSILON))
    if degenerated_triangle:
        return GjkState.NO_CONTACT, None, 0

    AO = -A[0]

    AB = B[0] - A[0]
    AC = C[0] - A[0]
    ABC = np.cross(AB, AC)

    if np.dot(np.cross(ABC, AC), AO) > -EPSILON:
        if np.dot(AC, AO) > -EPSILON:
            _set_point(v, v1, v2, 1, *A)
            n_points = 2
            search_direction = _triple_cross(AC, AO, AC)
        else:
            n_points, search_direction = _triangle_ab(A, B, AB, AO, v, v1, v2)
    else:
        if np.dot(np.cross(AB, ABC), AO) > -EPSILON:
            n_points, search_direction = _triangle_ab(A, B, AB, AO, v, v1, v2)
        else:
            if np.dot(ABC, AO) > -EPSILON:
                n_points = 3
                search_direction = ABC
            else:
                _set_point(v, v1, v2, 0, *B)
                _set_point(v, v1, v2, 1, *C)
                n_points = 3
                search_direction = -ABC

    return GjkState.CONTINUE, search_direction, n_points


@numba.njit(cache=True)
def _triangle_ab(A, B, AB, AO, v, v1, v2):
    if np.dot(AB, AO) > -EPSILON:
        _set_point(v, v1, v2, 0, *B)
        _set_point(v, v1, v2, 1, *A)
        n_points = 2
        search_direction = _triple_cross(AB, AO, AB)
    else:
        _set_point(v, v1, v2, 0, *A)
        n_points = 1
        search_direction = AO
    return n_points, search_direction


@numba.njit(cache=True)
def _tetrahedron(v, v1, v2):
    A = np.copy(v[3]), np.copy(v1[3]), np.copy(v1[3])
    B = np.copy(v[2]), np.copy(v1[2]), np.copy(v1[2])
    C = np.copy(v[1]), np.copy(v1[1]), np.copy(v2[1])
    D = np.copy(v[0]), np.copy(v1[0]), np.copy(v1[0])

    degenerated_tetrahedron = abs(point_to_triangle(
        A[0], np.row_stack((B[0], C[0], D[0])))[0]) < EPSILON_SQRT
    if degenerated_tetrahedron:
        return GjkState.NO_CONTACT, None, 0

    origin = np.zeros(3)
    origin_lies_on_tetrahedrons_face = (
        point_to_triangle(origin, np.row_stack((A[0], B[0], C[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((A[0], C[0], D[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((A[0], B[0], D[0])))[0] < EPSILON_SQRT
        or point_to_triangle(origin, np.row_stack((B[0], C[0], D[0])))[0] < EPSILON_SQRT
    )
    if origin_lies_on_tetrahedrons_face:
        return GjkState.CONTACT, None, 0

    # compute AO, AB, AC, AD segments and ABC, ACD, ADB normal vectors
    AO = -A[0]
    AB = B[0] - A[0]
    AC = C[0] - A[0]
    AD = D[0] - A[0]
    ABC = np.cross(AB, AC)
    ACD = np.cross(AC, AD)
    ADB = np.cross(AD, AB)

    # side (positive or negative) of B, C, D relative to planes ACD, ADB
    # and ABC respectively
    B_on_ACD = np.sign(np.dot(ACD, AB))
    C_on_ADB = np.sign(np.dot(ADB, AC))
    D_on_ABC = np.sign(np.dot(ABC, AD))

    # whether origin is on same side of ACD, ADB, ABC as B, C, D
    # respectively
    AB_O = np.sign(np.dot(ACD, AO)) == B_on_ACD
    AC_O = np.sign(np.dot(ADB, AO)) == C_on_ADB
    AD_O = np.sign(np.dot(ABC, AO)) == D_on_ABC

    origin_is_in_tetrahedron = AB_O and AC_O and AD_O
    if origin_is_in_tetrahedron:
        return GjkState.CONTACT, None, 0

    _rearrange_simplex_to_triangle(A, B, C, D, AB_O, AC_O, v, v1, v2)

    return _triangle(v, v1, v2)


@numba.njit(cache=True)
def _rearrange_simplex_to_triangle(A, B, C, D, AB_O, AC_O, v, v1, v2):
    if not AB_O:
        _set_point(v, v1, v2, 2, *A)
    elif not AC_O:
        _set_point(v, v1, v2, 1, *D)
        _set_point(v, v1, v2, 0, *B)
        _set_point(v, v1, v2, 2, *A)
    else:
        _set_point(v, v1, v2, 0, *C)
        _set_point(v, v1, v2, 1, *B)
        _set_point(v, v1, v2, 2, *A)


@numba.njit(cache=True)
def _triple_cross(a, b, c):
    """d = a x b x c"""
    return np.cross(np.cross(a, b), c)



"""Colliders used for collision detection with GJK and MPR algorithms."""
import abc
import numpy as np
from .geometry import (
    support_function_capsule, support_function_cylinder,
    convert_box_to_vertices, support_function_ellipsoid,
    support_function_sphere, support_function_cone, support_function_disk)
from .containment import (
    axis_aligned_bounding_box, sphere_aabb, box_aabb, cylinder_aabb,
    capsule_aabb, ellipsoid_aabb, cone_aabb, disk_aabb)
from .mesh import MeshHillClimbingSupportFunction
from .utils import plane_basis_from_normal
from aabbtree import AABB


class ConvexCollider(abc.ABC):
    """Convex collider base class.

    Parameters
    ----------
    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.

    Attributes
    ----------
    artist_ : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, artist=None):
        self.artist_ = artist

    @abc.abstractmethod
    def make_artist(self, c=None):
        """Make artist that represents this collider.

        Parameters
        ----------
        c : array-like, shape (3,), optional (default: None)
            Color of artist.
        """

    @abc.abstractmethod
    def first_vertex(self):
        """Get vertex with index 0 from collider to initialize GJK algorithm.

        Returns
        -------
        vertex : array, shape (3,)
            Vertex from collider.
        """

    @abc.abstractmethod
    def support_function(self, search_direction):
        """Support function for collider.

        Parameters
        ----------
        search_direction : array, shape (3,)
            Direction in which we search for extreme point of the collider.

        Returns
        -------
        support_point : array, shape (3,)
            Extreme point along search direction.
        """

    @abc.abstractmethod
    def center(self):
        """Returns the (approximate) center of the collider.

        Returns
        -------
        center : array, shape (3,)
            Center of the collider.
        """

    @abc.abstractmethod
    def update_pose(self, pose):
        """Update pose of collider.

        Parameters
        ----------
        pose : array, shape (4, 4)
            New pose of the collider.
        """

    @abc.abstractmethod
    def aabb(self):
        """Get axis-aligned bounding box.

        Returns
        -------
        aabb : AABB
            Axis-aligned bounding box.
        """


class ConvexHullVertices(ConvexCollider):
    """Convex hull of a set of vertices.

    Parameters
    ----------
    vertices : array, shape (n_vertices, 3)
        Vertices of the convex shape.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Artist for visualizer.
    """
    def __init__(self, vertices, artist=None):
        super(ConvexHullVertices, self).__init__(artist)
        self.vertices = vertices

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.PointCollection3D(self.vertices, s=0.005, c=c)

    def first_vertex(self):
        return self.vertices[0]

    def support_function(self, search_direction):
        return self.vertices[np.argmax(self.vertices.dot(search_direction))]

    def center(self):
        return np.mean(self.vertices, axis=0)

    def update_pose(self, pose):
        raise NotImplementedError("update_pose is not implemented!")

    def aabb(self):
        mins, maxs = axis_aligned_bounding_box(self.vertices)
        return AABB(np.array([mins, maxs]).T)


class Box(ConvexHullVertices):
    """Box collider.

    Parameters
    ----------
    box2origin : array, shape (4, 4)
        Pose of the box.

    size : array, shape (3,)
        Sizes of the box along its axes.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, box2origin, size, artist=None):
        super(Box, self).__init__(
            convert_box_to_vertices(box2origin, size), artist)
        self.box2origin = box2origin
        self.size = size

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.Box(size=self.size, A2B=self.box2origin, c=c)

    def center(self):
        return self.box2origin[:3, 3]

    def update_pose(self, pose):
        self.box2origin = pose
        self.vertices = convert_box_to_vertices(pose, self.size)
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = box_aabb(self.box2origin, self.size)
        return AABB(np.array([mins, maxs]).T)


class MeshGraph(ConvexCollider):
    """Mesh collider that uses triangles for hill climbing.

    Parameters
    ----------
    mesh2origin : array, shape (4, 4)
        Pose of the mesh.

    vertices : array, shape (n_vertices, 3)
        Vertices of the mesh.

    triangles : array, shape (n_triangles, 3)
        Indices of vertices that form triangles of the mesh.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, mesh2origin, vertices, triangles, artist=None):
        super(MeshGraph, self).__init__(artist)
        self.mesh2origin = mesh2origin
        self.vertices = vertices
        self.triangles = triangles
        self._support_function = MeshHillClimbingSupportFunction(
            mesh2origin, vertices, triangles)

    def make_artist(self, c=None):
        from .visualization import Mesh
        self.artist_ = Mesh(
            self.mesh2origin, self.vertices, self.triangles, c=c)

    def first_vertex(self):
        return self.mesh2origin[:3, 3] + np.dot(
            self.mesh2origin[:3, :3], self.vertices[0])

    def support_function(self, search_direction):
        return self._support_function(search_direction)[1]

    def center(self):
        return self.mesh2origin[:3, 3] + np.dot(
            self.mesh2origin[:3, :3], np.mean(self.vertices, axis=0))

    def update_pose(self, mesh2origin):
        self.mesh2origin = mesh2origin
        self._support_function.update_pose(mesh2origin)
        if self.artist_ is not None:
            self.artist_.set_data(mesh2origin)

    def aabb(self):
        mins, maxs = axis_aligned_bounding_box(
            self.mesh2origin[np.newaxis, :3, 3] + np.dot(
                self.vertices, self.mesh2origin[:3, :3].T))
        return AABB(np.array([mins, maxs]).T)


class Sphere(ConvexCollider):
    """Sphere collider.

    Parameters
    ----------
    center : array, shape (3,)
        Center of the sphere.

    radius : float
        Radius of the sphere.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, center, radius, artist=None):
        super(Sphere, self).__init__(artist)
        self.c = center
        self.radius = radius

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        sphere2origin = np.eye(4)
        sphere2origin[:3, 3] = self.c
        self.artist_ = pv.Sphere(radius=self.radius, A2B=sphere2origin, c=c)

    def center(self):
        return self.c

    def first_vertex(self):
        return self.c + np.array([0, 0, self.radius], dtype=float)

    def support_function(self, search_direction):
        return support_function_sphere(
            search_direction, np.ascontiguousarray(self.c), self.radius)

    def update_pose(self, pose):
        self.c = pose[:3, 3]
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = sphere_aabb(self.c, self.radius)
        return AABB(np.array([mins, maxs]).T)


class Capsule(ConvexCollider):
    """Capsule collider.

    Parameters
    ----------
    capsule2origin : array, shape (4, 4)
        Pose of the capsule.

    radius : float
        Radius of the capsule.

    height : float
        Height of the capsule.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, capsule2origin, radius, height, artist=None):
        super(Capsule, self).__init__(artist)
        self.capsule2origin = capsule2origin
        self.radius = radius
        self.height = height

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.Capsule(
            height=self.height, radius=self.radius, A2B=self.capsule2origin,
            c=c)

    def center(self):
        return self.capsule2origin[:3, 3]

    def first_vertex(self):
        return self.capsule2origin[:3, 3] - (
            self.radius + 0.5 * self.height) * self.capsule2origin[:3, 2]

    def support_function(self, search_direction):
        return support_function_capsule(
            search_direction, self.capsule2origin, self.radius, self.height)

    def update_pose(self, pose):
        self.capsule2origin = pose
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = capsule_aabb(
            self.capsule2origin, self.radius, self.height)
        return AABB(np.array([mins, maxs]).T)


class Ellipsoid(ConvexCollider):
    """Ellipsoid collider.

    Parameters
    ----------
    ellipsoid2origin : array, shape (4, 4)
        Pose of the ellipsoid.

    radii : array, shape (3,)
        Radii of the ellipsoid.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, ellipsoid2origin, radii, artist=None):
        super(Ellipsoid, self).__init__(artist)
        self.ellipsoid2origin = ellipsoid2origin
        self.radii = radii

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.Ellipsoid(
            radii=self.radii, A2B=self.ellipsoid2origin, c=c)

    def center(self):
        return self.ellipsoid2origin[:3, 3]

    def first_vertex(self):
        return (self.ellipsoid2origin[:3, 3]
                + self.radii[2] * self.ellipsoid2origin[:3, 2])

    def support_function(self, search_direction):
        return support_function_ellipsoid(
            search_direction, self.ellipsoid2origin, self.radii)

    def update_pose(self, pose):
        self.ellipsoid2origin = pose
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = ellipsoid_aabb(self.ellipsoid2origin, self.radii)
        return AABB(np.array([mins, maxs]).T)


class Cylinder(ConvexCollider):
    """Cylinder collider.

    Parameters
    ----------
    cylinder2origin : array, shape (4, 4)
        Pose of the cylinder.

    radius : float
        Radius of the cylinder.

    length : float
        Length of the cylinder.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, cylinder2origin, radius, length, artist=None):
        super(Cylinder, self).__init__(artist)
        self.cylinder2origin = cylinder2origin
        self.radius = radius
        self.length = length

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.Cylinder(
            length=self.length, radius=self.radius, A2B=self.cylinder2origin,
            c=c)

    def center(self):
        return self.cylinder2origin[:3, 3]

    def first_vertex(self):
        return (self.cylinder2origin[:3, 3]
                + 0.5 * self.length * self.cylinder2origin[:3, 2])

    def support_function(self, search_direction):
        return support_function_cylinder(
            search_direction, self.cylinder2origin, self.radius, self.length)

    def update_pose(self, pose):
        self.cylinder2origin = pose
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = cylinder_aabb(
            self.cylinder2origin, self.radius, self.length)
        return AABB(np.array([mins, maxs]).T)


class Disk(ConvexCollider):
    """Disk collider.

    Parameters
    ----------
    center : array, shape (3,)
        Center of the disk.

    radius : float
        Radius of the disk.

    normal : array, shape (3,)
        Normal to the plane in which the disk lies.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, center, radius, normal, artist=None):
        super(Disk, self).__init__(artist)
        self.c = center
        self.radius = radius
        self.normal = normal

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        x, y = plane_basis_from_normal(self.normal)
        disk2origin = np.eye(4)
        disk2origin[:3, :3] = np.column_stack((x, y, self.normal))
        disk2origin[:3, 3] = self.c
        self.artist_ = pv.Cylinder(
            A2B=disk2origin, radius=self.radius, length=0.01 * self.radius,
            c=c)

    def center(self):
        return self.c

    def first_vertex(self):
        x, _ = plane_basis_from_normal(self.normal)
        return self.c + self.radius * x

    def support_function(self, search_direction):
        return support_function_disk(
            search_direction, self.c, self.radius, self.normal)

    def update_pose(self, pose):
        self.c = pose[:3, 3]
        self.normal = pose[:3, 2]
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = disk_aabb(self.c, self.radius, self.normal)
        return AABB(np.array([mins, maxs]).T)


class Cone(ConvexCollider):
    """Cone collider.

    Parameters
    ----------
    cone2origin : array, shape (4, 4)
        Pose of the cone.

    radius : float
        Radius of the cone.

    height : float
        Length of the cone.

    artist : pytransform3d.visualizer.Artist, optional (default: None)
        Corresponding artist for visualizer.
    """
    def __init__(self, cone2origin, radius, height, artist=None):
        super(Cone, self).__init__(artist)
        self.cone2origin = cone2origin
        self.radius = radius
        self.height = height

    def make_artist(self, c=None):
        import pytransform3d.visualizer as pv
        self.artist_ = pv.Cone(
            A2B=self.cone2origin, radius=self.radius, height=self.height, c=c)

    def center(self):
        return (self.cone2origin[:3, 3]
                + 0.5 * self.height * self.cone2origin[:3, 2])

    def first_vertex(self):
        return (self.cone2origin[:3, 3]
                + self.height * self.cone2origin[:3, 2])

    def support_function(self, search_direction):
        return support_function_cone(
            search_direction, self.cone2origin, self.radius, self.height)

    def update_pose(self, pose):
        self.cone2origin = pose
        if self.artist_ is not None:
            self.artist_.set_data(pose)

    def aabb(self):
        mins, maxs = cone_aabb(
            self.cone2origin, self.radius, self.height)
        return AABB(np.array([mins, maxs]).T)
